# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import collections
import contextlib
import itertools
import queue
import threading

from debugpy.common import json, log, messaging, timestamp

from tests.patterns import some


SINGLE_LINE_REPR_LIMIT = 120
"""If repr() of an expectation or an occurrence is longer than this value, it will
be formatted to use multiple shorter lines if possible.
"""

# For use by Expectation.__repr__. Uses format() to create unique instances.
_INDENT = "{0}".format("_INDENT")
_DEDENT = "{0}".format("_DEDENT")


class Timeline(object):
    def __init__(self, name=None):
        self.name = str(name if name is not None else id(self))
        self.ignore_unobserved = []

        self._listeners = []  # [(expectation, callable)]
        self._index_iter = itertools.count(1)
        self._accepting_new = threading.Event()
        self._finalized = threading.Event()
        self._recorded_new = threading.Condition()
        self._record_queue = queue.Queue()

        self._recorder_thread = threading.Thread(
            target=self._recorder_worker, name=f"{self} recorder"
        )
        self._recorder_thread.daemon = True
        self._recorder_thread.start()

        # Set up initial environment for our first mark()
        self._last = None
        self._beginning = None
        self._accepting_new.set()

        self._beginning = self.mark("START")
        assert self._last is self._beginning
        self._proceeding_from = self._beginning

    def expect_frozen(self):
        if not self.is_frozen:
            raise Exception("Timeline can only be inspected while frozen.")

    def __iter__(self):
        self.expect_frozen()
        return self._beginning.and_following()

    def __len__(self):
        return len(self.history())

    @property
    def beginning(self):
        return self._beginning

    @property
    def last(self):
        self.expect_frozen()
        return self._last

    def history(self):
        self.expect_frozen()
        return list(iter(self))

    def __contains__(self, expectation):
        self.expect_frozen()
        return any(expectation.test(self.beginning, self.last))

    def all_occurrences_of(self, expectation):
        return tuple(occ for occ in self if occ == expectation)

    def __getitem__(self, index):
        assert isinstance(index, slice)
        assert index.step is None
        return Interval(self, index.start, index.stop)

    def __reversed__(self):
        self.expect_frozen()
        return self.last.and_preceding()

    @property
    def is_frozen(self):
        return not self._accepting_new.is_set()

    def freeze(self):
        self._accepting_new.clear()

    def unfreeze(self):
        if not self.is_final:
            self._accepting_new.set()

    @contextlib.contextmanager
    def frozen(self):
        was_frozen = self.is_frozen
        self.freeze()
        yield
        if not was_frozen and not self.is_final:
            self.unfreeze()

    @contextlib.contextmanager
    def unfrozen(self):
        was_frozen = self.is_frozen
        self.unfreeze()
        yield
        if was_frozen or self.is_final:
            self.freeze()

    @property
    def is_final(self):
        return self._finalized.is_set()

    def finalize(self):
        if self.is_final:
            return

        log.info("Finalizing timeline...")
        with self.unfrozen():
            self.mark("FINISH")

        with self.unfrozen():
            self._finalized.set()
            # Drain the record queue.
            self._record_queue.join()
            # Tell the recorder to shut itself down.
            self._record_queue.put(None)
            self._recorder_thread.join()

        assert self._record_queue.empty(), "Finalized timeline had pending records"

    def close(self):
        self.finalize()
        self[:].expect_no_unobserved()

    def __enter__(self):
        return self

    def __leave__(self, *args):
        self.close()

    def observe(self, *occurrences):
        for occ in occurrences:
            occ.observed = True

    def observe_all(self, expectation=None):
        self.expect_frozen()
        occs = (
            list(self)
            if expectation is None
            else [occ for occ in self if occ == expectation]
        )
        self.observe(*occs)

    def wait_until(self, condition, freeze=None):
        freeze = freeze or self.is_frozen
        try:
            with self._recorded_new:
                # First, test the condition against the timeline as it currently is.
                with self.frozen():
                    result = condition()
                    if result:
                        return result

                # Now keep spinning waiting for new occurrences to come, and test the
                # condition against every new batch in turn.
                self.unfreeze()
                while True:
                    self._recorded_new.wait()
                    with self.frozen():
                        result = condition()
                        if result:
                            return result
                    assert not self.is_final

        finally:
            if freeze:
                self.freeze()

    def _wait_until_realized(
        self, expectation, freeze=None, explain=True, observe=True
    ):
        def has_been_realized():
            for reasons in expectation.test(self.beginning, self.last):
                if observe:
                    self.expect_realized(expectation, explain=explain, observe=observe)
                return reasons

        reasons = self.wait_until(has_been_realized, freeze)
        return latest_of(reasons.values())

    def wait_until_realized(self, expectation, freeze=None, explain=True, observe=True):
        if explain:
            log.info("Waiting for {0!r}", expectation)
        return self._wait_until_realized(expectation, freeze, explain, observe)

    def wait_for(self, expectation, freeze=None, explain=True):
        assert expectation.has_lower_bound, (
            "Expectation must have a lower time bound to be used with wait_for()! "
            "Use >> to sequence an expectation against an occurrence to establish a lower bound, "
            "or wait_for_next() to wait for the next expectation since the timeline was last "
            "frozen, or wait_until_realized() when a lower bound is really not necessary."
        )
        if explain:
            log.info("Waiting for {0!r}", expectation)
        return self._wait_until_realized(expectation, freeze, explain=explain)

    def wait_for_next(self, expectation, freeze=True, explain=True, observe=True):
        if explain:
            log.info("Waiting for next {0!r}", expectation)
        return self._wait_until_realized(
            self._proceeding_from >> expectation, freeze, explain, observe
        )

    def new(self):
        self.expect_frozen()
        first_new = self._proceeding_from.next
        if first_new is not None:
            return self[first_new:]
        else:
            return self[self.last : self.last]

    def proceed(self):
        self.expect_frozen()
        self.new().expect_no_unobserved()
        self._proceeding_from = self.last
        self.unfreeze()

    def _expect_realized(self, expectation, first, explain=True, observe=True):
        self.expect_frozen()

        try:
            reasons = next(expectation.test(first, self.last))
        except StopIteration:
            log.info("No matching {0!r}", expectation)
            occurrences = list(first.and_following())
            log.info("Occurrences considered: {0!r}", occurrences)
            raise AssertionError("Expectation not matched")

        occs = tuple(reasons.values())
        assert occs
        if observe:
            self.observe(*occs)
        if explain:
            self._explain_how_realized(expectation, reasons)
        return occs if len(occs) > 1 else occs[0]

    def expect_realized(self, expectation, explain=True, observe=True):
        return self._expect_realized(expectation, self.beginning, explain, observe)

    def expect_new(self, expectation, explain=True, observe=True):
        assert (
            self._proceeding_from.next is not None
        ), "No new occurrences since last proceed()"
        return self._expect_realized(
            expectation, self._proceeding_from.next, explain, observe
        )

    def expect_not_realized(self, expectation):
        self.expect_frozen()
        assert expectation not in self

    def expect_no_new(self, expectation):
        self.expect_frozen()
        assert expectation not in self.new()

    def _explain_how_realized(self, expectation, reasons):
        message = f"Realized {expectation!r}"

        # For the breakdown, we want to skip any expectations that were exact occurrences,
        # since there's no point explaining that occurrence was realized by itself.
        skip = [exp for exp in reasons.keys() if isinstance(exp, Occurrence)]
        for exp in skip:
            reasons.pop(exp, None)

        if reasons == {expectation: some.object}:
            # If there's only one expectation left to explain, and it's the top-level
            # one, then we have already printed it, so just add the explanation.
            reason = reasons[expectation]
            if "\n" in message:
                message += f" == {reason!r}"
            else:
                message += f"\n      == {reason!r}"
        elif reasons:
            # Otherwise, break it down expectation by expectation.
            message += ":"
            for exp, reason in reasons.items():
                message += f"\n\n   where {exp!r}\n      == {reason!r}"
        else:
            message += "."

        log.info("{0}", message)

    def _record(self, occurrence, block=True):
        assert isinstance(occurrence, Occurrence)
        assert occurrence.timeline is None
        assert occurrence.timestamp is None
        assert (
            not self.is_final
        ), "Trying to record a new occurrence in a finalized timeline"

        self._record_queue.put(occurrence, block=block)
        if block:
            self._record_queue.join()

        return occurrence

    def _recorder_worker(self):
        while True:
            occ = self._record_queue.get()
            if occ is None:
                self._record_queue.task_done()
                break

            self._accepting_new.wait()
            with self._recorded_new:
                occ.timeline = self
                occ.timestamp = timestamp.current()
                occ.index = next(self._index_iter)

                if self._last is None:
                    self._beginning = occ
                    self._last = occ
                else:
                    assert self._last.timestamp <= occ.timestamp
                    occ.previous = self._last
                    self._last._next = occ
                    self._last = occ

                self._recorded_new.notify_all()
                self._record_queue.task_done()

                for exp, callback in tuple(self._listeners):
                    if exp == occ:
                        callback(occ)

    def mark(self, id, block=True):
        occ = Occurrence("mark", id)
        occ.id = id
        occ.observed = True
        return self._record(occ, block)

    def record_event(self, message, block=True):
        occ = EventOccurrence(message)
        return self._record(occ, block)

    def record_request(self, message, block=True):
        occ = RequestOccurrence(message)
        occ.observed = True
        return self._record(occ, block)

    def record_response(self, request_occ, message, block=True):
        occ = ResponseOccurrence(request_occ, message)
        return self._record(occ, block)

    def when(self, expectation, callback):
        """For every occurrence recorded after this call, invokes callback(occurrence)
        if occurrence == expectation.
        """
        self._listeners.append((expectation, callback))

    def _snapshot(self):
        last = self._last
        occ = self._beginning
        while True:
            yield occ
            if occ is last:
                break
            occ = occ._next

    def __repr__(self):
        return "|" + " >> ".join(repr(occ) for occ in self._snapshot()) + "|"

    def __str__(self):
        return "Timeline-" + self.name


class Interval(tuple):
    def __new__(cls, timeline, start, stop):
        assert start is None or isinstance(start, Expectation)
        assert stop is None or isinstance(stop, Expectation)
        if not isinstance(stop, Occurrence):
            timeline.expect_frozen()

        occs = ()
        if start is None:
            start = timeline._beginning

        for occ in start.and_following(up_to=stop):
            if occ == stop:
                break
            if occ == start:
                occs = occ.and_following(up_to=stop)
                break

        result = super().__new__(cls, occs)
        result.timeline = timeline
        result.start = start
        result.stop = stop
        return result

    def __contains__(self, expectation):
        return any(expectation.test(self[0], self[-1])) if len(self) > 0 else False

    def all_occurrences_of(self, expectation):
        return tuple(occ for occ in self if occ == expectation)

    def expect_no_unobserved(self):
        if not self:
            return

        unobserved = [
            occ
            for occ in self
            if not occ.observed
            and all(exp != occ for exp in self.timeline.ignore_unobserved)
        ]
        if not unobserved:
            return

        raise log.error(
            "Unobserved occurrences detected:\n\n{0}\n\nignoring unobserved:\n\n{1}",
            "\n\n".join(repr(occ) for occ in unobserved),
            "\n\n".join(repr(exp) for exp in self.timeline.ignore_unobserved),
        )


class Expectation(object):
    timeline = None
    has_lower_bound = False

    def test(self, first, last):
        raise NotImplementedError()

    def wait(self, freeze=None, explain=True):
        assert (
            self.timeline is not None
        ), "Expectation must be bound to a timeline to be waited on."
        return self.timeline.wait_for(self, freeze, explain)

    def wait_until_realized(self, freeze=None):
        return self.timeline.wait_until_realized(self, freeze)

    def __eq__(self, other):
        if self is other:
            return True
        elif isinstance(other, Occurrence) and any(self.test(other, other)):
            return True
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self == other

    def after(self, other):
        return SequencedExpectation(other, self)

    def when(self, condition):
        return ConditionalExpectation(self, condition)

    def __rshift__(self, other):
        return self if other is None else other.after(self)

    def __and__(self, other):
        assert isinstance(other, Expectation)
        return AndExpectation(self, other)

    def __or__(self, other):
        assert isinstance(other, Expectation)
        return OrExpectation(self, other)

    def __xor__(self, other):
        assert isinstance(other, Expectation)
        return XorExpectation(self, other)

    def __hash__(self):
        return hash(id(self))

    def __repr__(self):
        raise NotImplementedError


class DerivativeExpectation(Expectation):
    def __init__(self, *expectations):
        self.expectations = expectations
        assert len(expectations) > 0
        assert all(isinstance(exp, Expectation) for exp in expectations)

        timelines = {id(exp.timeline): exp.timeline for exp in expectations}
        timelines.pop(id(None), None)
        if len(timelines) > 1:
            offending_expectations = ""
            for tl_id, tl in timelines.items():
                offending_expectations += f"\n    {tl_id}: {tl!r}\n"
            raise log.error(
                "Cannot mix expectations from multiple timelines:\n{0}",
                offending_expectations,
            )
        for tl in timelines.values():
            self.timeline = tl

    @property
    def has_lower_bound(self):
        return all(exp.has_lower_bound for exp in self.expectations)

    def flatten(self):
        """Flattens nested expectation chains.

        If self is of type E, and given an expectation like::

            E(E(e1, E(e2, e3)), E(E(e4, e5), E(e6)))

        flatten() produces an iterator over::

            e1, e2, e3, e4, e5, e6
        """
        for exp in self.expectations:
            if type(exp) is type(self):
                for exp in exp.flatten():
                    yield exp
            else:
                yield exp

    def describe(self, newline):
        """Returns an iterator describing this expectation. This method is used
        to implement repr().

        For every yielded _INDENT and _DEDENT, a newline and the appropriate amount
        of spaces for correct indentation at the current level is added to the repr.

        For every yielded Expectation, describe() is invoked recursively.

        For every other yielded value, str(value) added to the repr.

        newline is set to either "" or "\n", depending on whether the repr must be
        single-line or multiline. Implementations of describe() should use it in
        lieu of raw "\n" insofar as possible; however, repr() will automatically
        fall back to multiline mode if "\n" occurs in single-line mode.

        The default implementation produces a description that looks like::

            (e1 OP e2 OP e3 OP ...)

        where OP is the value of self.OPERATOR.
        """
        op = self.OPERATOR

        yield "("
        yield _INDENT

        first = True
        for exp in self.flatten():
            if first:
                first = False
            else:
                yield " " + op + " "
                yield newline
            yield exp

        yield _DEDENT
        yield ")"

    def __repr__(self):
        def indent():
            return indent.level * "    " if newline else ""

        def recurse(exp):
            for item in exp.describe(newline):
                if isinstance(item, DerivativeExpectation):
                    recurse(item)
                elif item is _INDENT:
                    indent.level += 1
                    result.append(newline + indent())
                elif item is _DEDENT:
                    assert indent.level > 0, "_DEDENT without matching _INDENT"
                    indent.level -= 1
                    result.append(newline + indent())
                else:
                    item = str(item).replace("\n", "\n" + indent())
                    result.append(item)

        # Try single-line repr first.
        indent.level = 0
        newline = ""
        result = []
        recurse(self)
        s = "".join(result)
        if len(s) <= SINGLE_LINE_REPR_LIMIT and "\n" not in s:
            return s

        # If it was too long, or had newlines anyway, fall back to multiline.
        assert indent.level == 0
        newline = "\n"
        result[:] = []
        recurse(self)
        return "".join(result)


class SequencedExpectation(DerivativeExpectation):
    OPERATOR = ">>"

    def __init__(self, first, second):
        super().__init__(first, second)

    @property
    def first(self):
        return self.expectations[0]

    @property
    def second(self):
        return self.expectations[1]

    def test(self, first, last):
        assert isinstance(first, Occurrence)
        assert isinstance(last, Occurrence)

        for first_reasons in self.first.test(first, last):
            first_occs = first_reasons.values()
            lower_bound = latest_of(first_occs).next
            if lower_bound is not None:
                for second_reasons in self.second.test(lower_bound, last):
                    reasons = second_reasons.copy()
                    reasons.update(first_reasons)
                    yield reasons

    @property
    def has_lower_bound(self):
        return self.first.has_lower_bound or self.second.has_lower_bound


class OrExpectation(DerivativeExpectation):
    OPERATOR = "|"

    def test(self, first, last):
        assert isinstance(first, Occurrence)
        assert isinstance(last, Occurrence)

        for exp in self.expectations:
            for reasons in exp.test(first, last):
                yield reasons

    def __or__(self, other):
        assert isinstance(other, Expectation)
        return OrExpectation(*(self.expectations + (other,)))


class AndExpectation(DerivativeExpectation):
    OPERATOR = "&"

    def test(self, first, last):
        assert isinstance(first, Occurrence)
        assert isinstance(last, Occurrence)

        if len(self.expectations) <= 1:
            for exp in self.expectations:
                for reasons in exp.test(first, last):
                    yield reasons
            return

        lhs = self.expectations[0]
        rhs = AndExpectation(*self.expectations[1:])
        for lhs_reasons in lhs.test(first, last):
            for rhs_reasons in rhs.test(first, last):
                reasons = lhs_reasons.copy()
                reasons.update(rhs_reasons)
                yield reasons

    @property
    def has_lower_bound(self):
        return any(exp.has_lower_bound for exp in self.expectations)

    def __and__(self, other):
        assert isinstance(other, Expectation)
        return AndExpectation(*(self.expectations + (other,)))

    def __repr__(self):
        return "(" + " & ".join(repr(exp) for exp in self.expectations) + ")"


class XorExpectation(DerivativeExpectation):
    OPERATOR = "^"

    def test(self, first, last):
        assert isinstance(first, Occurrence)
        assert isinstance(last, Occurrence)

        reasons = None
        for exp in self.expectations:
            for exp_reasons in exp.test(first, last):
                if reasons is None:
                    reasons = exp_reasons
                else:
                    return

        if reasons is not None:
            yield reasons

    @property
    def has_lower_bound(self):
        return all(exp.has_lower_bound for exp in self.expectations)

    def __xor__(self, other):
        assert isinstance(other, Expectation)
        return XorExpectation(*(self.expectations + (other,)))


class ConditionalExpectation(DerivativeExpectation):
    def __init__(self, expectation, condition):
        self.condition = condition
        super().__init__(expectation)

    @property
    def expectation(self):
        return self.expectations[0]

    def test(self, first, last):
        assert isinstance(first, Occurrence)
        assert isinstance(last, Occurrence)

        for reasons in self.expectation.test(first, last):
            occs = reasons.values()
            if self.condition(*occs):
                yield reasons

    def describe(self, newline):
        yield "?"
        yield self.expectation


class PatternExpectation(Expectation):
    def __init__(self, *circumstances):
        self.circumstances = circumstances

    def test(self, first, last):
        assert isinstance(first, Occurrence)
        assert isinstance(last, Occurrence)

        for occ in first.and_following(up_to=last, inclusive=True):
            if occ.circumstances == self.circumstances:
                yield {self: occ}

    def describe(self):
        rest = repr(self.circumstances[1:])
        if rest.endswith(",)"):
            rest = rest[:-2] + ")"
        return f"<{self.circumstances[0]}{rest}>"

    def __repr__(self):
        return self.describe()


def Mark(id):
    return PatternExpectation("mark", id)


def _describe_message(message_type, *items):
    items = (("type", message_type),) + items
    d = collections.OrderedDict(items)

    # Keep it all on one line if it's short enough, but indent longer ones.
    for format_string in "{0:indent=None}", "{0}":
        s = format_string.format(json.repr(d))
        s = "{..., " + s[1:]

        # Used by some.dict.containing to inject ... as needed.
        s = s.replace('"\\u0002...": "...\\u0003"', "...")
        # Used by some.* and by Event/Request/Response expectations below.
        s = s.replace('"\\u0002', "")
        s = s.replace('\\u0003"', "")

        if len(s) <= SINGLE_LINE_REPR_LIMIT:
            break

    return s


def Event(event, body=some.object):
    exp = PatternExpectation("event", event, body)
    items = (("event", event),)
    if body is some.object:
        items += (("\002...", "...\003"),)
    else:
        items += (("body", body),)
    exp.describe = lambda: _describe_message("event", *items)
    return exp


def Request(command, arguments=some.object):
    exp = PatternExpectation("request", command, arguments)
    items = (("command", command),)
    if arguments is some.object:
        items += (("\002...", "...\003"),)
    else:
        items += (("arguments", arguments),)
    exp.describe = lambda: _describe_message("request", *items)
    return exp


def Response(request, body=some.object):
    assert isinstance(request, Expectation) or isinstance(request, RequestOccurrence)

    exp = PatternExpectation("response", request, body)
    exp.timeline = request.timeline
    exp.has_lower_bound = request.has_lower_bound

    # Try to be as specific as possible.
    if isinstance(request, Expectation):
        if request.circumstances[0] != "request":
            exp.describe = lambda: f"response to {request!r}"
            return
        else:
            items = (("command", request.circumstances[1]),)
    else:
        items = (("command", request.command),)

    if isinstance(request, Occurrence):
        items += (("request_seq", request.seq),)

    if body is some.object:
        items += (("\002...", "...\003"),)
    elif body is some.error or body == some.error:
        items += (("success", False),)
        if body == some.error:
            items += (("message", str(body)),)
    else:
        items += (("body", body),)

    exp.describe = lambda: _describe_message("response", *items)
    return exp


class Occurrence(Expectation):
    has_lower_bound = True

    def __init__(self, *circumstances):
        assert circumstances
        self.circumstances = circumstances

        self.timeline = None
        self.timestamp = None
        self.index = None
        self.previous = None
        self._next = None
        self.observed = False

    @property
    def next(self):
        if self.timeline is None:
            return None

        with self.timeline.frozen():
            was_last = self is self.timeline.last
            occ = self._next

        if was_last:
            # The .next property of the last occurrence in a timeline can change
            # at any moment when timeline isn't frozen. So if it wasn't frozen by
            # the caller, this was an unsafe operation, and we should complain.
            self.timeline.expect_frozen()

        return occ

    def preceding(self):
        it = self.and_preceding()
        next(it)
        return it

    def and_preceding(self, up_to=None, inclusive=False):
        assert self.timeline is not None
        assert up_to is None or isinstance(up_to, Expectation)

        if isinstance(up_to, Occurrence) and self < up_to:
            return

        occ = self
        while occ != up_to:
            yield occ
            occ = occ.previous

        if inclusive:
            yield occ

    def following(self):
        it = self.and_following()
        next(it)
        return it

    def and_following(self, up_to=None, inclusive=False):
        assert self.timeline is not None
        assert up_to is None or isinstance(up_to, Expectation)
        if up_to is None:
            self.timeline.expect_frozen()

        if isinstance(up_to, Occurrence) and up_to < self:
            return

        occ = self
        while occ != up_to:
            yield occ
            occ = occ.next

        if inclusive:
            yield occ

    def precedes(self, occurrence):
        assert isinstance(occurrence, Occurrence)
        return any(occ is self for occ in occurrence.preceding())

    def follows(self, occurrence):
        return occurrence.precedes(self)

    def realizes(self, expectation):
        assert isinstance(expectation, Expectation)
        return expectation == self

    def test(self, first, last):
        assert isinstance(first, Occurrence)
        assert isinstance(last, Occurrence)

        for occ in first.and_following(up_to=last, inclusive=True):
            if occ is self:
                yield {self: self}

    def __lt__(self, occurrence):
        return self.precedes(occurrence)

    def __gt__(self, occurrence):
        return occurrence.precedes(self)

    def __le__(self, occurrence):
        return self is occurrence or self < occurrence

    def __ge__(self, occurrence):
        return self is occurrence or self > occurrence

    def __rshift__(self, expectation):
        assert isinstance(expectation, Expectation)
        return expectation.after(self)

    def __hash__(self):
        return hash(id(self))

    def __repr__(self):
        return (
            "" if self.observed else "*"
        ) + f"{self.index}.{self.describe_circumstances()}"

    def describe_circumstances(self):
        rest = repr(self.circumstances[1:])
        if rest.endswith(",)"):
            rest = rest[:-2] + ")"
        return f"{self.circumstances[0]}{rest}"


class MessageOccurrence(Occurrence):
    """An occurrence representing a DAP message (event, request, or response).

    Its circumstances == (self.TYPE, self._key, self._data).
    """

    TYPE = None
    """Used for self.circumstances[0].

    Must be defined by subclasses.
    """

    def __init__(self, message):
        assert self.TYPE
        assert isinstance(message, messaging.Message)

        # Assign message first for the benefit of self._data in child classes.
        self.message = message
        super().__init__(self.TYPE, self._key, self._data)

    @property
    def seq(self):
        return self.message.seq

    @property
    def _key(self):
        """The part of the message that describes it in general terms - e.g. for
        an event, it's the name of the event.
        """
        raise NotImplementedError

    @property
    def _data(self):
        """The part of the message that is used for matching expectations, excluding
        self._key.
        """
        raise NotImplementedError

    @property
    def _id(self):
        """The part of the message that is necessary and sufficient to uniquely
        identify it. Used for __repr__().

        Must be an ordered list of key-value tuples, suitable for OrderedDict().
        """
        return [("seq", self.message.seq), ("type", self.TYPE)]

    def __call__(self, *args, **kwargs):
        return self.message(*args, **kwargs)

    def describe_circumstances(self):
        id = collections.OrderedDict(self._id)

        # Keep it all on one line if it's short enough, but indent longer ones.
        s = f"{json.repr(id):indent=None}"
        if len(s) > SINGLE_LINE_REPR_LIMIT:
            s = f"{json.repr(id)}"
        return s

    # For messages, we don't want to include their index, because they already have
    # "seq" to identify them uniquely, and including both is confusing.
    def __repr__(self):
        return ("" if self.observed else "*") + self.describe_circumstances()


class EventOccurrence(MessageOccurrence):
    TYPE = "event"

    def __init__(self, message):
        assert isinstance(message, messaging.Event)
        super().__init__(message)

    @property
    def event(self):
        return self.message.event

    @property
    def body(self):
        return self.message.body

    @property
    def _key(self):
        return self.event

    @property
    def _data(self):
        return self.body

    @property
    def _id(self):
        return super()._id + [("event", self.message.event)]


class RequestOccurrence(MessageOccurrence):
    TYPE = "request"

    def __init__(self, message):
        assert isinstance(message, messaging.Request)
        super().__init__(message)
        self.response = None
        if isinstance(message, messaging.OutgoingRequest):
            self.on_response = message.on_response

    @property
    def command(self):
        return self.message.command

    @property
    def arguments(self):
        return self.message.arguments

    @property
    def _key(self):
        return self.command

    @property
    def _data(self):
        return self.arguments

    @property
    def _id(self):
        return super()._id + [("command", self.message.command)]

    def wait_for_response(self, freeze=True, raise_if_failed=True):
        response = Response(self, some.object).wait_until_realized(freeze)
        assert response.observed
        if raise_if_failed and not response.success:
            raise response.body
        else:
            return response


class ResponseOccurrence(MessageOccurrence):
    TYPE = "response"

    def __init__(self, request_occ, message):
        assert isinstance(request_occ, RequestOccurrence)
        assert isinstance(message, messaging.Response)

        # Assign request first for the benefit of self._key.
        self.request = request_occ
        request_occ.response = self
        super().__init__(message)

    @property
    def body(self):
        return self.message.body

    @property
    def result(self):
        return self.message.result

    @property
    def success(self):
        return self.message.success

    @property
    def _key(self):
        return self.request

    @property
    def _data(self):
        return self.body

    @property
    def _id(self):
        return super()._id + [
            ("command", self.message.request.command),
            ("request_seq", self.message.request.seq),
        ]

    def causing(self, *expectations):
        for exp in expectations:
            (self >> exp).wait()
        return self


def earliest_of(occurrences):
    return min(occurrences, key=lambda occ: occ.index)


def latest_of(occurrences):
    return max(occurrences, key=lambda occ: occ.index)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Imported from test code that runs under debugpy, and allows that code
to communcate back to the test. Works in conjunction with debug_session
fixture and its backchannel method."""

__all__ = ["port", "receive", "send"]

import atexit
import os
import socket

import debuggee
from debugpy.common import log, messaging


def send(value):
    _stream.write_json(value)


def receive():
    return _stream.read_json()


def close():
    global _socket, _stream
    if _socket is None:
        return

    log.info("Shutting down {0}...", name)
    try:
        _socket.shutdown(socket.SHUT_RDWR)
    except Exception:
        pass
    finally:
        _socket = None
        try:
            _stream.close()
        except Exception:
            pass
        finally:
            _stream = None


class _stream:
    def _error(*_):
        raise AssertionError("Backchannel is not set up for this process")

    read_json = write_json = _error
    close = lambda: None


name = f"backchannel-{debuggee.session_id}"
port = os.environ.pop("DEBUGPY_TEST_BACKCHANNEL_PORT", None)
if port is not None:
    port = int(port)
    log.info("Connecting {0} to port {1}...", name, port)

    _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        _socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        _socket.connect(("localhost", port))
    except Exception:
        _socket.close()
        raise
    else:
        _stream = messaging.JsonIOStream.from_socket(  # noqa
            _socket, name="backchannel"
        )
        atexit.register(close)



# Note that await in the top-level isn't valid in general, but we compile
# it specifically accepting it, so, that's ok.
import asyncio

await asyncio.sleep(.01)
a = 1  # Break here
await asyncio.sleep(.01)
b = 2
await asyncio.sleep(.01)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from tests import debug
from tests.patterns import some
from tests.timeline import Event


@pytest.mark.parametrize("module", ["module", ""])
@pytest.mark.parametrize("line", ["line", ""])
def test_stack_format(pyfile, target, run, module, line):
    @pyfile
    def code_to_debug():
        import debuggee
        from test_module import do_something

        debuggee.setup()
        do_something()

    @pyfile
    def test_module():
        def do_something():
            print("break here")  # @bp

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.set_breakpoints(test_module, all)

        stop = session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(test_module, line="bp")]
        )
        stack_trace = session.request(
            "stackTrace",
            {
                "threadId": stop.thread_id,
                "format": {"module": bool(module), "line": bool(line)},
            },
        )
        assert stack_trace["totalFrames"] > 0
        name = stack_trace["stackFrames"][0]["name"]
        if line:
            assert (": " + str(test_module.lines["bp"])) in name
        if module:
            assert "test_module" in name
        session.request_continue()


def test_module_events(pyfile, target, run):
    @pyfile
    def module2():
        def do_more_things():
            print("done")  # @bp

    @pyfile
    def module1():
        import module2

        def do_something():
            module2.do_more_things()

    @pyfile
    def test_code():
        import debuggee

        debuggee.setup()

        from module1 import do_something

        do_something()

    with debug.Session() as session:
        with run(session, target(test_code)):
            session.set_breakpoints(module2, all)

        session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(module2, line="bp")]
        )

        # Stack trace after the stop will trigger module events, but they are only
        # sent after the trace response, so we need to wait for them separately.
        # The order isn't guaranteed, either, so just wait for any 3 modules.
        session.timeline.wait_until_realized(
            Event("module") >> Event("module") >> Event("module")
        )
        modules = {
            event.body["module"]["name"]: event.body["module"]["path"]
            for event in session.all_occurrences_of(Event("module"))
        }
        assert modules == some.dict.containing(
            {
                "__main__": some.path(test_code),
                "module1": some.path(module1),
                "module2": some.path(module2),
            }
        )

        session.request_continue()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import py
import os

from debugpy.common import json
from tests.patterns import some


class Target(object):
    """Describes Python code that gets run by a Runner."""

    def __init__(self, filename, args=()):
        if filename is not None and not isinstance(filename, py.path.local):
            filename = py.path.local(filename)

        self.filename = filename
        self.args = args

        if self.filename is None:
            self.code = None
        else:
            with open(self.filename.strpath, "rb") as f:
                self.code = f.read().decode("utf-8")

    def configure(self, session):
        """Configures the session to execute this target.

        This should only modify session.config, but gets access to the entire session
        to retrieve information about it.
        """

        raise NotImplementedError

    def cli(self, env):
        """Provides the command line arguments, suitable for passing to python or
        python -m debugpy, to execute this target.

        Returns command line arguments as a list, e.g. ["-m", "module"].

        If any environment variables are needed to properly interpret the command
        line - e.g. PYTHONPATH - the implementation should send them in env.
        """
        raise NotImplementedError

    @property
    def argslist(self):
        args = self.args
        if isinstance(args, str):
            return [args]
        else:
            return list(args)

    @property
    def co_filename(self):
        """co_filename of code objects created at runtime from the source that this
        Target describes, assuming no path mapping.
        """
        assert (
            self.filename is not None
        ), "co_filename requires Target created from filename"
        return self.filename.strpath

    @property
    def source(self):
        """DAP "source" JSON for this Target."""
        return some.dap.source(py.path.local(self.co_filename))

    @property
    def lines(self):
        """Same as self.filename.lines, if it is valid - e.g. for @pyfile objects."""
        assert (
            self.filename is not None
        ), "lines() requires Target created from filename"
        return self.filename.lines


class Program(Target):
    """
    A Python script, executed directly.

    By default executes a program through its absolute path with:
        python /full/path/to/foo.py

    If made relative through make_relative(cwd), the cwd is set and the
    launch is made relative to it.

        i.e.:
        Given a path such as /full/path/to/foo.py
        if make_relative('/full/path') is called,
        the cwd becomes `/full/path` and the launch is made as:
            python to/foo.py
    """

    pytest_id = "program"

    def __init__(self, filename, args=()):
        super().__init__(filename, args)
        self._cwd = ""

    def make_relative(self, cwd):
        if not cwd.endswith(("/", "\\")):
            cwd += os.path.sep
        assert self.filename.strpath.lower().startswith(cwd.lower())
        self._cwd = cwd

    def _get_relative_program(self):
        assert self._cwd
        relative_filename = self.filename.strpath[len(self._cwd) :]
        assert not relative_filename.startswith(("/", "\\"))
        return relative_filename

    def __repr__(self):
        if self._cwd:
            return f"program (relative) {json.repr(self._cwd)} / {json.repr(self._get_relative_program())}"
        else:
            return f"program {json.repr(self.filename.strpath)}"

    def configure(self, session):
        if self._cwd:
            session.config["cwd"] = self._cwd
            session.config["program"] = self._get_relative_program()
        else:
            session.config["program"] = self.filename.strpath

        session.config["args"] = self.args

    def cli(self, env):
        if self._cwd:
            cli = [self._get_relative_program()]
        else:
            cli = [self.filename.strpath]
        cli += self.argslist
        return cli


class Module(Target):
    """A Python module, executed by name: python -m foo.bar

    If created from a filename, the module name is the name of the file, and the
    Target will automatically add a PYTHONPATH entry.
    """

    pytest_id = "module"

    def __init__(self, filename=None, name=None, args=()):
        assert (filename is None) ^ (name is None)
        super().__init__(filename, args)
        self.name = name if name is not None else self.filename.purebasename

    def __repr__(self):
        return f"module {self.name}"

    def configure(self, session):
        session.config["module"] = self.name
        session.config["args"] = self.args

    def cli(self, env):
        if self.filename is not None:
            env.prepend_to("PYTHONPATH", self.filename.dirname)
        return ["-m", self.name] + self.argslist


class Code(Target):
    """A snippet of Python code: python -c "print('foo')"

    If created from a filename, the code is the contents of the file.
    """

    pytest_id = "code"

    def __init__(self, filename=None, code=None, args=()):
        assert (filename is None) ^ (code is None)
        super().__init__(filename, args)
        if code is not None:
            self.code = code

    def __repr__(self):
        lines = self.code.split("\n")
        return f"code: {json.repr(lines)}"

    def configure(self, session):
        session.config["code"] = self.code
        session.config["args"] = self.args

    def cli(self, env):
        return ["-c", self.code] + self.argslist

    @property
    def co_filename(self):
        return "<string>"

    @property
    def source(self):
        """DAP "source" JSON for this Target."""
        return some.dap.source("<string>")


all_named = [Program, Module]
"""All targets that produce uniquely named code objects at runtime, and thus can
have breakpoints set in them.
"""

all_unnamed = [Code]
"""All targets that produce unnamed code objects at runtime, and thus cannot have
breakpoints set in them.
"""

all = all_named + all_unnamed



# -*- coding: utf-8 -*-
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
import re
import sys

import tests
from tests import debug, test_data, timeline
from tests.debug import runners, targets
from tests.patterns import some

bp_root = test_data / "bp"


if not tests.full:

    @pytest.fixture(params=[runners.launch, runners.attach_connect["cli"]])
    def run(request):
        return request.param


@pytest.mark.parametrize("target", targets.all_named)
def test_path_with_ampersand(target, run):
    test_py = bp_root / "a&b" / "test.py"

    with debug.Session() as session:
        with run(session, target(test_py)):
            session.set_breakpoints(test_py, ["two"])

        session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(test_py, line="two")]
        )
        session.request_continue()


@pytest.mark.parametrize("target", targets.all_named)
def test_path_with_unicode(target, run):
    test_py = bp_root / "ನನ್ನ_ಸ್ಕ್ರಿಪ್ಟ್.py"

    with debug.Session() as session:
        with run(session, target(test_py)):
            session.set_breakpoints(test_py, ["bp"])

        session.wait_for_stop(
            "breakpoint",
            expected_frames=[some.dap.frame(test_py, name="ಏನಾದರೂ_ಮಾಡು", line="bp")],
        )
        session.request_continue()


conditions = {
    ("condition", "i==5"): lambda i: i == 5,
    ("hitCondition", "5"): lambda i: i == 5,
    ("hitCondition", "==5"): lambda i: i == 5,
    ("hitCondition", ">5"): lambda i: i > 5,
    ("hitCondition", ">=5"): lambda i: i >= 5,
    ("hitCondition", "<5"): lambda i: i < 5,
    ("hitCondition", "<=5"): lambda i: i <= 5,
    ("hitCondition", "%3"): lambda i: i % 3 == 0,
}


@pytest.mark.parametrize("condition_kind, condition", list(conditions.keys()))
@pytest.mark.parametrize("target", targets.all_named)
def test_conditional_breakpoint(pyfile, target, run, condition_kind, condition):
    hit = conditions[condition_kind, condition]

    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        for i in range(1, 10):
            print(i)  # @bp

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.request(
                "setBreakpoints",
                {
                    "source": {"path": code_to_debug},
                    "breakpoints": [
                        {"line": code_to_debug.lines["bp"], condition_kind: condition}
                    ],
                },
            )

        for i in range(1, 10):
            if not hit(i):
                continue
            session.wait_for_stop(
                expected_frames=[some.dap.frame(code_to_debug, line="bp")]
            )
            var_i = session.get_variable("i")
            assert var_i == some.dict.containing(
                {"name": "i", "evaluateName": "i", "type": "int", "value": str(i)}
            )
            session.request_continue()


def test_crossfile_breakpoint(pyfile, target, run):
    @pyfile
    def script1():
        import debuggee

        debuggee.setup()

        def do_something():
            print("do something")  # @bp

    @pyfile
    def script2():
        import debuggee
        import script1

        debuggee.setup()
        script1.do_something()  # @bp
        print("Done")

    with debug.Session() as session:
        with run(session, target(script2)):
            session.set_breakpoints(script1, all)
            session.set_breakpoints(script2, all)

        session.wait_for_stop(expected_frames=[some.dap.frame(script2, line="bp")])
        session.request_continue()

        session.wait_for_stop(expected_frames=[some.dap.frame(script1, line="bp")])
        session.request_continue()


# NameError in condition is a special case: pydevd is configured to skip traceback for
# name errors. See https://github.com/microsoft/ptvsd/issues/853 for more details. For
# all other errors, we should be printing traceback.
@pytest.mark.parametrize("error_name", ["NameError", ""])
def test_error_in_condition(pyfile, target, run, error_name):
    error_name = error_name or "ZeroDivisionError"

    condition, expect_traceback = {
        "NameError": ("no_such_name", False),
        "ZeroDivisionError": ("1 / 0", True),
    }[error_name]

    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        for i in range(1, 10):  # @bp
            pass

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.request(
                "setBreakpoints",
                {
                    "source": {"path": code_to_debug},
                    "breakpoints": [
                        {"line": code_to_debug.lines["bp"], "condition": condition}
                    ],
                },
            )
    occurrences = session.timeline.all_occurrences_of(
        timeline.Event("output", some.dict.containing({"category": "important"})),
    )
    
    if expect_traceback:
        assert len(occurrences) == 10
        for occurrence in occurrences:
            assert error_name in occurrence.body['output']
    else:
        assert len(occurrences) == 0


@pytest.mark.parametrize("condition", ["condition", ""])
@pytest.mark.parametrize("target", targets.all_named)
def test_log_point(pyfile, target, run, condition):
    @pyfile
    def code_to_debug():
        import debuggee
        import sys

        debuggee.setup()
        for i in range(0, 10):
            sys.stderr.write(str(i * 10) + "\n")  # @bp
            sys.stderr.flush()
        x = 4  # @wait_for_output

    lines = code_to_debug.lines
    with debug.Session() as session:
        session.config["redirectOutput"] = True

        with run(session, target(code_to_debug)):
            bp = {"line": lines["bp"], "logMessage": "{i}"}
            if condition:
                bp["condition"] = "i == 5"
            session.request(
                "setBreakpoints",
                {
                    "source": {"path": code_to_debug},
                    "breakpoints": [bp, {"line": lines["wait_for_output"]}],
                },
            )

        session.wait_for_stop(
            "breakpoint",
            expected_frames=[some.dap.frame(code_to_debug, line="wait_for_output")],
        )
        session.request_continue()

    # print() should produce both actual output, and "output" events on stderr,
    # but logpoints should only produce "output" events on stdout.
    if "internalConsole" not in str(run):
        assert not session.captured_stdout()

    if condition:
        expected_stdout = "5\r?\n"
    else:
        expected_stdout = "".join(
            (r"{0}\r?\n".format(re.escape(str(i))) for i in range(0, 10))
        )
    expected_stderr = "".join(
        (r"{0}\r?\n".format(re.escape(str(i * 10))) for i in range(0, 10))
    )
    assert session.output("stdout") == some.str.matching(expected_stdout)
    assert session.output("stderr") == some.str.matching(expected_stderr)


@pytest.mark.parametrize("run", [runners.launch])
def test_breakpoint_in_package_main(run):
    testpkgs = test_data / "testpkgs"
    main_py = testpkgs / "pkg1" / "__main__.py"

    with debug.Session() as session:
        session.expected_exit_code = 42
        session.config["cwd"] = testpkgs.strpath

        with run(session, targets.Module(name="pkg1")):
            session.set_breakpoints(main_py, ["two"])

        session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(main_py, line="two")]
        )
        session.request_continue()


def test_add_and_remove_breakpoint(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        for i in range(0, 10):
            print(i)  # @bp
        x = 4  # @wait_for_output

    with debug.Session() as session:
        session.config["redirectOutput"] = True

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(code_to_debug, line="bp")]
        )

        # Remove breakpoint inside the loop.
        session.set_breakpoints(code_to_debug, ["wait_for_output"])
        session.request_continue()

        session.wait_for_stop(
            "breakpoint",
            expected_frames=[some.dap.frame(code_to_debug, line="wait_for_output")],
        )
        session.request_continue()

    expected_stdout = "".join(f"{i}\n" for i in range(0, 10))
    assert session.output("stdout") == expected_stdout


def test_breakpoint_in_nonexistent_file(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            breakpoints = session.set_breakpoints("nonexistent_file.py", [1])
            assert breakpoints == [
                {
                    "verified": False,
                    "message": "Breakpoint in file that does not exist.",
                    "source": some.dict.containing(
                        {"path": some.path("nonexistent_file.py")}
                    ),
                    "line": 1,
                    "id": 0,
                }
            ]


@pytest.mark.flaky(retries=2, delay=1)
def test_invalid_breakpoints(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        # For markers below, rN = requested breakpoint, eN = expected breakpoint.
        # If there's no eN for some rN, it's assumed to be the same line.
        # fmt: off
        b = True
        while b:         # @e0-37,e0-38,e0-39
            pass         # @r0
            break

        print()         # @e1-37 
        [               # @r1,e2
            1, 2, 3,    # @e2-37,e2-38
        ]               # @r2

        print()         # @e3,e4
        print(1,        # @r3
              2, 3,     # @r4
              4, 5, 6)
        # fmt: on

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            requested_markers = ["r" + str(i) for i in range(0, 3)]

            bps = session.set_breakpoints(code_to_debug, requested_markers)
            actual_lines = [bp["line"] for bp in bps]

            expected_markers = []
            for r in requested_markers:
                e_generic = "e" + r[1:]
                e_versioned = (
                    e_generic
                    + "-"
                    + str(sys.version_info.major)
                    + str(sys.version_info.minor)
                )
                for e in e_versioned, e_generic, r:
                    if e in code_to_debug.lines:
                        expected_markers.append(e)
                        break

            expected_lines = [
                code_to_debug.lines[marker] for marker in expected_markers
            ]
            assert actual_lines == expected_lines

        # Now let's make sure that we hit all of the expected breakpoints,
        # and stop where we expect them to be.

        # If there's multiple breakpoints on the same line, we only stop once,
        # so remove duplicates first.
        expected_lines = sorted(set(expected_lines))
        if (3, 8) <= sys.version_info < (3, 9):
            # We'll actually hit @e2-38 first, and only then @r1, because there's
            # a line event for [ when the list creation is finished on 3.8).
            # See https://bugs.python.org/issue38508 for details.
            expected_lines[1], expected_lines[2] = expected_lines[2], expected_lines[1]

        while expected_lines:
            expected_line = expected_lines.pop(0)
            session.wait_for_stop(
                "breakpoint",
                expected_frames=[some.dap.frame(code_to_debug, line=expected_line)],
            )
            session.request_continue()


def test_deep_stacks(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        def deep_stack(level):
            if level <= 0:
                print("done")  # @bp
                return level
            deep_stack(level - 1)

        deep_stack(100)

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()
        assert len(stop.frames) > 100

        # Now try to retrieve the same stack in chunks, and check that it matches.
        frames = []
        for _ in range(5):
            stack_trace = session.request(
                "stackTrace",
                {"threadId": stop.thread_id, "startFrame": len(frames), "levels": 25},
            )

            assert stack_trace["totalFrames"] > 0
            frames += stack_trace["stackFrames"]

        assert stop.frames == frames
        session.request_continue()


@pytest.mark.parametrize("target", targets.all)
@pytest.mark.parametrize("func", ["breakpoint", "debugpy.breakpoint"])
def test_break_api(pyfile, target, run, func):
    @pyfile
    def code_to_debug():
        import debuggee
        import debugpy  # noqa
        import sys

        debuggee.setup()
        func = eval(sys.argv[1])
        func()
        print("break here")  # @break

    with debug.Session() as session:
        target = target(code_to_debug, args=[func])
        with run(session, target):
            pass

        session.wait_for_stop(
            expected_frames=[some.dap.frame(target.source, target.lines["break"])]
        )
        session.request_continue()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
from tests import debug
from tests.patterns import some
from _pydevd_bundle.pydevd_constants import IS_PY313_OR_GREATER

@pytest.mark.skipif(IS_PY313_OR_GREATER, reason="gevent is not up to date with 3.13 (_tstate_lock is not part of thread anymore)")
def test_gevent(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import os

        os.environ["GEVENT_SUPPORT"] = "True"

        from gevent import monkey

        monkey.patch_all()

        import threading
        from time import sleep

        # Note: importing ptvsd after the `monkey.patch_all` (with the GEVENT_SUPPORT
        # flag == 'True').
        import debuggee  # noqa

        debuggee.setup()

        called = []

        def myfunc():
            class MyGreenletThread(threading.Thread):
                def run(self):
                    print("break here", self.name)  # @bp
                    for _i in range(5):
                        called.append(self.name)  # break here
                        sleep()

            t1 = MyGreenletThread()
            t1.name = "t1"
            t2 = MyGreenletThread()
            t2.name = "t2"

            t1.start()
            t2.start()

            for t1 in (t1, t2):
                t1.join()

            # With gevent it's always the same (gevent coroutine support makes thread
            # switching serial).
            expected = ["t1", "t1", "t2", "t1", "t2", "t1", "t2", "t1", "t2", "t2"]
            if called != expected:
                raise AssertionError("Expected:\n%s\nFound:\n%s" % (expected, called))

        myfunc()

    with debug.Session() as session:
        session.config["gevent"] = True

        if str(run).startswith("attach"):
            session.spawn_debuggee.env["GEVENT_SUPPORT"] = "True"

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(code_to_debug, "bp")]
        )
        session.request_continue()

        session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(code_to_debug, "bp")]
        )

        session.request_continue()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import os
import pytest
import sys

from tests import debug
from tests.debug import runners


@pytest.mark.parametrize("new_value", [None, "42"])
@pytest.mark.parametrize("case", ["match_case", "mismatch_case"])
@pytest.mark.parametrize("run", runners.all_launch)
def test_env_replace_var(pyfile, target, run, case, new_value):
    @pyfile
    def code_to_debug():
        import os

        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send(dict(os.environ))

    varname = "DEBUGPY_DUMMY_ENV_VAR"

    with debug.Session() as session:
        backchannel = session.open_backchannel()
        session.config.env[
            varname if case == "match_case" else varname.lower()
        ] = new_value

        os.environ[varname] = "1"
        try:
            with run(session, target(code_to_debug)):
                pass
        finally:
            del os.environ[varname]

        env = backchannel.receive()
        if case == "match_case":
            # If case matches, debug config should replace global env var regardless
            # of the platform.
            if new_value is None:
                assert varname not in env
            else:
                assert env[varname] == "42"
        elif sys.platform == "win32":
            # On Win32, variable names are case-insensitive, so debug config should
            # replace the global env var even if there is a case mismatch.
            if new_value is None:
                assert varname not in env
            else:
                assert env[varname] == "42"
            assert varname.lower() not in env
        else:
            # On other platforms, variable names are case-sensitive, so case mismatch
            # should result in two different variables.
            assert env[varname] == "1"
            if new_value is None:
                assert varname.lower() not in env
            else:
                assert env[varname.lower()] == "42"



import sys
from _pydev_bundle._pydev_saved_modules import thread
from _pydevd_bundle.pydevd_constants import IS_PY313_OR_GREATER
import pycompletionserver
import socket
from urllib.parse import quote_plus

start_new_thread = thread.start_new_thread

BUILTIN_MOD = "builtins"


def send(s, msg):
    s.send(bytearray(msg, "utf-8"))


import unittest


class TestCPython(unittest.TestCase):
    def test_message(self):
        t = pycompletionserver.CompletionServer(0)

        l = []
        l.append(("Def", "description", "args"))
        l.append(("Def1", "description1", "args1"))
        l.append(("Def2", "description2", "args2"))

        msg = t.processor.format_completion_message(None, l)

        self.assertEqual("@@COMPLETIONS(None,(Def,description,args),(Def1,description1,args1),(Def2,description2,args2))END@@", msg)
        l = []
        l.append(("Def", "desc,,r,,i()ption", ""))
        l.append(("Def(1", "descriptio(n1", ""))
        l.append(("De,f)2", "de,s,c,ription2", ""))
        msg = t.processor.format_completion_message(None, l)
        self.assertEqual(
            "@@COMPLETIONS(None,(Def,desc%2C%2Cr%2C%2Ci%28%29ption, ),(Def%281,descriptio%28n1, ),(De%2Cf%292,de%2Cs%2Cc%2Cription2, ))END@@",
            msg,
        )

    def create_connections(self):
        """
        Creates the connections needed for testing.
        """
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((pycompletionserver.HOST, 0))
        server.listen(1)  # socket to receive messages.

        t = pycompletionserver.CompletionServer(server.getsockname()[1])
        t.exit_process_on_kill = False
        start_new_thread(t.run, ())

        s, _addr = server.accept()

        return t, s

    def read_msg(self):
        finish = False
        msg = ""
        while finish == False:
            m = self.socket.recv(1024 * 4)
            m = m.decode("utf-8")
            if m.startswith("@@PROCESSING"):
                sys.stdout.write("Status msg: %s\n" % (msg,))
            else:
                msg += m

            if msg.find("END@@") != -1:
                finish = True

        return msg

    @unittest.skipIf(IS_PY313_OR_GREATER and sys.platform == "linux",
                     "Flakey on Linux")
    def test_completion_sockets_and_messages(self):
        t, socket = self.create_connections()
        self.socket = socket

        try:
            # now that we have the connections all set up, check the code completion messages.
            msg = quote_plus("math")
            send(socket, "@@IMPORTS:%sEND@@" % msg)  # math completions
            completions = self.read_msg()
            # print_ unquote_plus(completions)

            # math is a builtin and because of that, it starts with None as a file
            start = "@@COMPLETIONS(None,(__doc__,"
            start_2 = "@@COMPLETIONS(None,(__name__,"
            if (
                "/math.so," in completions
                or "/math.cpython-33m.so," in completions
                or "/math.cpython-34m.so," in completions
                or "math.cpython-35m" in completions
                or "math.cpython-36m" in completions
                or "math.cpython-37m" in completions
                or "math.cpython-38" in completions
                or "math.cpython-39" in completions
                or "math.cpython-310" in completions
                or "math.cpython-311" in completions
                or "math.cpython-312" in completions
            ):
                return
            self.assertTrue(
                completions.startswith(start) or completions.startswith(start_2),
                "%s DOESNT START WITH %s" % (completions, (start, start_2)),
            )

            self.assertTrue("@@COMPLETIONS" in completions)
            self.assertTrue("END@@" in completions)

            # now, test i
            msg = quote_plus("%s.list" % BUILTIN_MOD)
            send(socket, "@@IMPORTS:%s\nEND@@" % msg)
            found = self.read_msg()
            self.assertTrue("sort" in found, "Could not find sort in: %s" % (found,))

            # now, test search
            msg = quote_plus("inspect.ismodule")
            send(socket, "@@SEARCH%sEND@@" % msg)  # math completions
            found = self.read_msg()
            self.assertTrue("inspect.py" in found)
            for i in range(33, 100):
                if str(i) in found:
                    break
            else:
                self.fail("Could not find the ismodule line in %s" % (found,))

            # now, test search
            msg = quote_plus("inspect.CO_NEWLOCALS")
            send(socket, "@@SEARCH%sEND@@" % msg)  # math completions
            found = self.read_msg()
            self.assertTrue("inspect.py" in found)
            self.assertTrue("CO_NEWLOCALS" in found)

            # now, test search
            msg = quote_plus("inspect.BlockFinder.tokeneater")
            send(socket, "@@SEARCH%sEND@@" % msg)
            found = self.read_msg()
            self.assertTrue("inspect.py" in found)
        #            self.assertTrue('CO_NEWLOCALS' in found)

        # reload modules test
        #        send(socket, '@@RELOAD_MODULES_END@@')
        #        ok = self.read_msg()
        #        self.assertEqual('@@MSG_OK_END@@' , ok)
        #        this test is not executed because it breaks our current enviroment.

        finally:
            try:
                sys.stdout.write("succedded...sending kill msg\n")
                self.send_kill_msg(socket)

                #                while not hasattr(t, 'ended'):
                #                    pass #wait until it receives the message and quits.

                socket.close()
                self.socket.close()
            except:
                pass

    def send_kill_msg(self, socket):
        socket.send(pycompletionserver.MSG_KILL_SERVER)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

# The actual patterns are defined here, so that tests.patterns.some can redefine
# builtin names like str, int etc without affecting the implementations in this
# file - some.* then provides shorthand aliases.

import collections
import itertools
import py.path
import re
import sys

from debugpy.common import util
import pydevd_file_utils


class Some(object):
    """A pattern that can be tested against a value with == to see if it matches."""

    def matches(self, value):
        raise NotImplementedError

    def __repr__(self):
        try:
            return self.name
        except AttributeError:
            raise NotImplementedError

    def __eq__(self, value):
        return self.matches(value)

    def __ne__(self, value):
        return not self.matches(value)

    def __invert__(self):
        """The inverse pattern - matches everything that this one doesn't."""
        return Not(self)

    def __or__(self, pattern):
        """Union pattern - matches if either of the two patterns match."""
        return Either(self, pattern)

    def such_that(self, condition):
        """Same pattern, but it only matches if condition() is true."""
        return SuchThat(self, condition)

    def in_range(self, start, stop):
        """Same pattern, but it only matches if the start <= value < stop."""
        return InRange(self, start, stop)

    def equal_to(self, obj):
        return EqualTo(self, obj)

    def not_equal_to(self, obj):
        return NotEqualTo(self, obj)

    def same_as(self, obj):
        return SameAs(self, obj)

    def matching(self, regex, flags=0):
        """Same pattern, but it only matches if re.match(regex, flags) produces
        a match that corresponds to the entire string.
        """
        return Matching(self, regex, flags)

    # Used to obtain the JSON representation for logging. This is a hack, because
    # JSON serialization doesn't allow to customize raw output - this function can
    # only substitute for another object that is normally JSON-serializable. But
    # for patterns, we want <...> in the logs, not'"<...>". Thus, we insert dummy
    # marker chars here, such that it looks like "\002<...>\003" in serialized JSON -
    # and then tests.timeline._describe_message does a string substitution on the
    # result to strip out '"\002' and '\003"'.
    def __getstate__(self):
        return "\002" + repr(self) + "\003"


class Not(Some):
    """Matches the inverse of the pattern."""

    def __init__(self, pattern):
        self.pattern = pattern

    def __repr__(self):
        return f"~{self.pattern!r}"

    def matches(self, value):
        return value != self.pattern


class Either(Some):
    """Matches either of the patterns."""

    def __init__(self, *patterns):
        assert len(patterns) > 0
        self.patterns = tuple(patterns)

    def __repr__(self):
        try:
            return self.name
        except AttributeError:
            return "({0})".format(" | ".join(repr(pat) for pat in self.patterns))

    def matches(self, value):
        return any(pattern == value for pattern in self.patterns)

    def __or__(self, pattern):
        return Either(*(self.patterns + (pattern,)))


class Object(Some):
    """Matches anything."""

    name = "<?>"

    def matches(self, value):
        return True


class Thing(Some):
    """Matches anything that is not None."""

    name = "<>"

    def matches(self, value):
        return value is not None


class InstanceOf(Some):
    """Matches any object that is an instance of the specified type."""

    def __init__(self, classinfo, name=None):
        if isinstance(classinfo, type):
            classinfo = (classinfo,)
        assert len(classinfo) > 0 and all(
            (isinstance(cls, type) for cls in classinfo)
        ), "classinfo must be a type or a tuple of types"

        self.name = name
        self.classinfo = classinfo

    def __repr__(self):
        if self.name:
            name = self.name
        else:
            name = " | ".join(cls.__name__ for cls in self.classinfo)
        return f"<{name}>"

    def matches(self, value):
        return isinstance(value, self.classinfo)


class Path(Some):
    """Matches any string that matches the specified path.

    Uses os.path.normcase() to normalize both strings before comparison.
    """

    def __init__(self, path):
        if isinstance(path, py.path.local):
            path = path.strpath
        assert isinstance(path, str)
        self.path = path

    def __repr__(self):
        return "path({self.path!r})"

    def __str__(self):
        return self.path

    def __getstate__(self):
        return self.path

    def matches(self, other):
        if isinstance(other, py.path.local):
            other = other.strpath

        if isinstance(other, str):
            pass
        elif isinstance(other, bytes):
            other = other.encode(sys.getfilesystemencoding())
        else:
            return NotImplemented

        left = pydevd_file_utils.get_path_with_real_case(self.path)
        right = pydevd_file_utils.get_path_with_real_case(other)
        return left == right


class ListContaining(Some):
    """Matches any list that contains the specified subsequence of elements."""

    def __init__(self, *items):
        self.items = tuple(items)

    def __repr__(self):
        if not self.items:
            return "[...]"
        s = repr(list(self.items))
        return f"[..., {s[1:-1]}, ...]"

    def __getstate__(self):
        items = ["\002...\003"]
        if not self.items:
            return items
        items *= 2
        items[1:1] = self.items
        return items

    def matches(self, other):
        if not isinstance(other, list):
            return NotImplemented

        items = self.items
        if not items:
            return True  # every list contains an empty sequence
        if len(items) == 1:
            return self.items[0] in other

        # Zip the other list with itself, shifting by one every time, to produce
        # tuples of equal length with items - i.e. all potential subsequences. So,
        # given other=[1, 2, 3, 4, 5] and items=(2, 3, 4), we want to get a list
        # like [(1, 2, 3), (2, 3, 4), (3, 4, 5)] - and then search for items in it.
        iters = [itertools.islice(other, i, None) for i in range(0, len(items))]
        subseqs = zip(*iters)
        return any(subseq == items for subseq in subseqs)


class DictContaining(Some):
    """Matches any dict that contains the specified key-value pairs::

    d1 = {'a': 1, 'b': 2, 'c': 3}
    d2 = {'a': 1, 'b': 2}
    assert d1 == some.dict.containing(d2)
    assert d2 != some.dict.containing(d1)
    """

    def __init__(self, items):
        self.items = collections.OrderedDict(items)

    def __repr__(self):
        return dict.__repr__(self.items)[:-1] + ", ...}"

    def __getstate__(self):
        items = self.items.copy()
        items["\002..."] = "...\003"
        return items

    def matches(self, other):
        if not isinstance(other, dict):
            return NotImplemented
        any = Object()
        d = {key: any for key in other}
        d.update(self.items)
        return d == other


class Also(Some):
    """Base class for patterns that narrow down another pattern."""

    def __init__(self, pattern):
        self.pattern = pattern

    def matches(self, value):
        return self.pattern == value and self._also(value)

    def _also(self, value):
        raise NotImplementedError


class SuchThat(Also):
    """Matches only if condition is true."""

    def __init__(self, pattern, condition):
        super().__init__(pattern)
        self.condition = condition

    def __repr__(self):
        try:
            return self.name
        except AttributeError:
            return f"({self.pattern!r} if {util.nameof(self.condition)})"

    def _also(self, value):
        return self.condition(value)


class InRange(Also):
    """Matches only if the value is within the specified range."""

    def __init__(self, pattern, start, stop):
        super().__init__(pattern)
        self.start = start
        self.stop = stop

    def __repr__(self):
        try:
            return self.name
        except AttributeError:
            return f"({self.start!r} <= {self.pattern!r} < {self.stop!r})"

    def _also(self, value):
        return self.start <= value < self.stop


class EqualTo(Also):
    """Matches any object that is equal to the specified object."""

    def __init__(self, pattern, obj):
        super().__init__(pattern)
        self.obj = obj

    def __repr__(self):
        return repr(self.obj)

    def __str__(self):
        return str(self.obj)

    def __getstate__(self):
        return self.obj

    def _also(self, value):
        return self.obj == value


class NotEqualTo(Also):
    """Matches any object that is not equal to the specified object."""

    def __init__(self, pattern, obj):
        super().__init__(pattern)
        self.obj = obj

    def __repr__(self):
        return f"<!={self.obj!r}>"

    def _also(self, value):
        return self.obj != value


class SameAs(Also):
    """Matches one specific object only (i.e. makes '==' behave like 'is')."""

    def __init__(self, pattern, obj):
        super().__init__(pattern)
        self.obj = obj

    def __repr__(self):
        return f"<is {self.obj!r}>"

    def _also(self, value):
        return self.obj is value


class Matching(Also):
    """Matches any string that matches the specified regular expression."""

    def __init__(self, pattern, regex, flags=0):
        assert isinstance(regex, bytes) or isinstance(regex, str)
        super().__init__(pattern)
        self.regex = regex
        self.flags = flags

    def __repr__(self):
        s = repr(self.regex)
        if s[0] in "bu":
            return s[0] + "/" + s[2:-1] + "/"
        else:
            return "/" + s[1:-1] + "/"

    def _also(self, value):
        regex = self.regex

        # re.match() always starts matching at the beginning, but does not require
        # a complete match of the string - append "$" to ensure the latter.
        if isinstance(regex, bytes):
            if not isinstance(value, bytes):
                return NotImplemented
            regex += b"$"
        elif isinstance(regex, str):
            if not isinstance(value, str):
                return NotImplemented
            regex += "$"
        else:
            raise AssertionError()

        return re.match(regex, value, self.flags) is not None



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import collections
import functools
import itertools
import pytest
import threading
import time

from debugpy.common import log, messaging
from tests.patterns import some
from tests.timeline import Timeline, Mark, Event, Request, Response


class MessageFactory(object):
    """A factory for DAP messages that are not bound to a message channel."""

    def __init__(self):
        self._seq_iter = itertools.count(1)
        self.messages = collections.OrderedDict()
        self.event = functools.partial(self._make, messaging.Event)
        self.request = functools.partial(self._make, messaging.Request)
        self.response = functools.partial(self._make, messaging.Response)

    def _make(self, message_type, *args, **kwargs):
        message = message_type(None, next(self._seq_iter), *args, **kwargs)
        self.messages[message.seq] = message
        return message


@pytest.fixture
def make_timeline(request):
    """Provides a timeline factory. All timelines created by this factory
    are automatically finalized and checked for basic consistency after the
    end of the test.
    """

    timelines = []

    def factory():
        timeline = Timeline()
        timelines.append(timeline)
        with timeline.frozen():
            assert timeline.beginning is not None
            initial_history = [some.object.same_as(timeline.beginning)]
            assert timeline.history() == initial_history
        return timeline, initial_history

    yield factory
    log.newline()

    try:
        failed = request.node.call_result.failed
    except AttributeError:
        pass
    else:
        if not failed:
            for timeline in timelines:
                timeline.finalize()
                history = timeline.history()
                history.sort(key=lambda occ: occ.timestamp)
                assert history == timeline.history()
                assert history[-1] == Mark("FINISH")
                timeline.close()


@pytest.mark.timeout(1)
def test_occurrences(make_timeline):
    timeline, initial_history = make_timeline()

    mark1 = timeline.mark("dum")
    mark2 = timeline.mark("dee")
    mark3 = timeline.mark("dum")

    assert mark1 == Mark("dum")
    assert mark1.id == "dum"
    assert mark2 == Mark("dee")
    assert mark2.id == "dee"
    assert mark3 == Mark("dum")
    assert mark3.id == "dum"

    with timeline.frozen():
        assert timeline.history() == initial_history + [
            some.object.same_as(mark1),
            some.object.same_as(mark2),
            some.object.same_as(mark3),
        ]
        timeline.finalize()

    assert timeline.all_occurrences_of(Mark("dum")) == (
        some.object.same_as(mark1),
        some.object.same_as(mark3),
    )
    assert timeline.all_occurrences_of(Mark("dee")) == (some.object.same_as(mark2),)

    assert timeline[:].all_occurrences_of(Mark("dum")) == (
        some.object.same_as(mark1),
        some.object.same_as(mark3),
    )

    # Lower boundary is inclusive.
    assert timeline[mark1:].all_occurrences_of(Mark("dum")) == (
        some.object.same_as(mark1),
        some.object.same_as(mark3),
    )
    assert timeline[mark2:].all_occurrences_of(Mark("dum")) == (
        some.object.same_as(mark3),
    )
    assert timeline[mark3:].all_occurrences_of(Mark("dum")) == (
        some.object.same_as(mark3),
    )

    # Upper boundary is exclusive.
    assert timeline[:mark1].all_occurrences_of(Mark("dum")) == ()
    assert timeline[:mark2].all_occurrences_of(Mark("dum")) == (
        some.object.same_as(mark1),
    )
    assert timeline[:mark3].all_occurrences_of(Mark("dum")) == (
        some.object.same_as(mark1),
    )


def expectation_of(message):
    if isinstance(message, messaging.Event):
        return Event(some.str.equal_to(message.event), some.dict.equal_to(message.body))
    elif isinstance(message, messaging.Request):
        return Request(
            some.str.equal_to(message.command), some.dict.equal_to(message.arguments)
        )
    else:
        raise AssertionError("Unsupported message type")


def test_event(make_timeline):
    timeline, initial_history = make_timeline()
    messages = MessageFactory()

    event_msg = messages.event("stopped", {"reason": "pause"})
    event_exp = expectation_of(event_msg)
    event = timeline.record_event(event_msg)

    assert event == event_exp
    assert event.message is event_msg
    assert event.circumstances == ("event", event_msg.event, event_msg.body)

    with timeline.frozen():
        assert timeline.last is event
        assert timeline.history() == initial_history + [some.object.same_as(event)]
        timeline.expect_realized(event_exp)


@pytest.mark.parametrize("outcome", ["success", "failure"])
def test_request_response(make_timeline, outcome):
    timeline, initial_history = make_timeline()
    messages = MessageFactory()

    request_msg = messages.request("next", {"threadId": 3})
    request_exp = expectation_of(request_msg)
    request = timeline.record_request(request_msg)

    assert request == request_exp
    assert request.command == request_msg.command
    assert request.arguments == request_msg.arguments
    assert request.circumstances == (
        "request",
        request_msg.command,
        request_msg.arguments,
    )

    with timeline.frozen():
        assert timeline.last is request
        assert timeline.history() == initial_history + [some.object.same_as(request)]
        timeline.expect_realized(request_exp)
        timeline.expect_realized(Request("next"))

    response_msg = messages.response(
        request_msg, {} if outcome == "success" else Exception("error!")
    )
    response = timeline.record_response(request, response_msg)
    assert response.request is request
    assert response.body == response_msg.body
    if outcome == "success":
        assert response.success
    else:
        assert not response.success
    assert response.circumstances == ("response", request, response_msg.body)

    assert response == Response(request, response_msg.body)
    assert response == Response(request, some.object)
    assert response == Response(request)

    if outcome == "success":
        assert response == Response(request, some.dict)
        assert response != Response(request, some.error)
    else:
        assert response != Response(request, some.dict)
        assert response == Response(request, some.error)

    assert response == Response(request_exp, response_msg.body)
    assert response == Response(request_exp, some.object)
    assert response == Response(request_exp)

    if outcome == "success":
        assert response == Response(request_exp, some.dict)
        assert response != Response(request_exp, some.error)
    else:
        assert response != Response(request_exp, some.dict)
        assert response == Response(request_exp, some.error)

    with timeline.frozen():
        assert timeline.last is response
        assert timeline.history() == initial_history + [
            some.object.same_as(request),
            some.object.same_as(response),
        ]

        timeline.expect_realized(Response(request))
        timeline.expect_realized(Response(request, response_msg.body))
        timeline.expect_realized(Response(request, some.object))

        if outcome == "success":
            timeline.expect_realized(Response(request, some.dict))
            timeline.expect_not_realized(Response(request, some.error))
        else:
            timeline.expect_not_realized(Response(request, some.dict))
            timeline.expect_realized(Response(request, some.error))

        timeline.expect_realized(Response(request))
        timeline.expect_realized(Response(request_exp, response_msg.body))
        timeline.expect_realized(Response(request_exp, some.object))

        if outcome == "success":
            timeline.expect_realized(Response(request_exp, some.dict))
            timeline.expect_not_realized(Response(request_exp, some.error))
        else:
            timeline.expect_not_realized(Response(request_exp, some.dict))
            timeline.expect_realized(Response(request_exp, some.error))


def test_after(make_timeline):
    timeline, _ = make_timeline()
    first = timeline.mark("first")

    second_exp = first >> Mark("second")
    with timeline.frozen():
        assert second_exp not in timeline

    timeline.mark("second")
    with timeline.frozen():
        assert second_exp in timeline

    timeline.mark("first")
    with timeline.frozen():
        assert Mark("second") >> Mark("first") in timeline


def test_before(make_timeline):
    timeline, _ = make_timeline()
    t = timeline.beginning

    first = timeline.mark("first")
    timeline.mark("second")

    with timeline.frozen():
        assert t >> Mark("second") >> Mark("first") not in timeline
        assert Mark("second") >> first not in timeline

    third = timeline.mark("third")

    with timeline.frozen():
        assert t >> Mark("second") >> Mark("first") not in timeline
        assert Mark("second") >> first not in timeline
        assert t >> Mark("second") >> Mark("third") in timeline
        assert Mark("second") >> third in timeline


def test_and(make_timeline):
    eggs_exp = Mark("eggs")
    ham_exp = Mark("ham")
    cheese_exp = Mark("cheese")

    timeline, _ = make_timeline()
    t = timeline.beginning

    with timeline.frozen():
        assert t >> (eggs_exp & ham_exp) not in timeline
        assert t >> (ham_exp & eggs_exp) not in timeline
        assert t >> (cheese_exp & ham_exp & eggs_exp) not in timeline

    timeline.mark("eggs")
    with timeline.frozen():
        assert t >> (eggs_exp & ham_exp) not in timeline
        assert t >> (ham_exp & eggs_exp) not in timeline
        assert t >> (cheese_exp & ham_exp & eggs_exp) not in timeline

    timeline.mark("ham")
    with timeline.frozen():
        assert t >> (eggs_exp & ham_exp) in timeline
        assert t >> (ham_exp & eggs_exp) in timeline
        assert t >> (cheese_exp & ham_exp & eggs_exp) not in timeline

    timeline.mark("cheese")
    with timeline.frozen():
        assert t >> (eggs_exp & ham_exp) in timeline
        assert t >> (ham_exp & eggs_exp) in timeline
        assert t >> (cheese_exp & ham_exp & eggs_exp) in timeline


def test_or(make_timeline):
    eggs_exp = Mark("eggs")
    ham_exp = Mark("ham")
    cheese_exp = Mark("cheese")

    timeline, _ = make_timeline()
    t = timeline.beginning

    with timeline.frozen():
        assert t >> (eggs_exp | ham_exp) not in timeline
        assert t >> (ham_exp | eggs_exp) not in timeline
        assert t >> (cheese_exp | ham_exp | eggs_exp) not in timeline

    timeline.mark("eggs")
    with timeline.frozen():
        assert t >> (eggs_exp | ham_exp) in timeline
        assert t >> (ham_exp | eggs_exp) in timeline
        assert t >> (cheese_exp | ham_exp | eggs_exp) in timeline

    timeline.mark("cheese")
    with timeline.frozen():
        assert t >> (eggs_exp | ham_exp) in timeline
        assert t >> (ham_exp | eggs_exp) in timeline
        assert t >> (cheese_exp | ham_exp | eggs_exp) in timeline

    timeline.mark("ham")
    with timeline.frozen():
        assert t >> (eggs_exp | ham_exp) in timeline
        assert t >> (ham_exp | eggs_exp) in timeline
        assert t >> (cheese_exp | ham_exp | eggs_exp) in timeline
        t = timeline.last

    timeline.mark("cheese")
    with timeline.frozen():
        assert t >> (eggs_exp | ham_exp) not in timeline
        assert t >> (ham_exp | eggs_exp) not in timeline
        assert t >> (cheese_exp | ham_exp | eggs_exp) in timeline


def test_xor(make_timeline):
    eggs_exp = Mark("eggs")
    ham_exp = Mark("ham")
    cheese_exp = Mark("cheese")

    timeline, _ = make_timeline()
    t1 = timeline.beginning

    with timeline.frozen():
        assert t1 >> (eggs_exp ^ ham_exp) not in timeline
        assert t1 >> (ham_exp ^ eggs_exp) not in timeline
        assert t1 >> (cheese_exp ^ ham_exp ^ eggs_exp) not in timeline

    timeline.mark("eggs")
    with timeline.frozen():
        assert t1 >> (eggs_exp ^ ham_exp) in timeline
        assert t1 >> (ham_exp ^ eggs_exp) in timeline
        assert t1 >> (cheese_exp ^ ham_exp ^ eggs_exp) in timeline
        t2 = timeline.last

    timeline.mark("ham")
    with timeline.frozen():
        assert t1 >> (eggs_exp ^ ham_exp) not in timeline
        assert t2 >> (eggs_exp ^ ham_exp) in timeline
        assert t1 >> (ham_exp ^ eggs_exp) not in timeline
        assert t2 >> (ham_exp ^ eggs_exp) in timeline
        assert t1 >> (cheese_exp ^ ham_exp ^ eggs_exp) not in timeline
        assert t2 >> (cheese_exp ^ ham_exp ^ eggs_exp) in timeline


def test_conditional(make_timeline):
    def is_exciting(occ):
        return occ == Event(some.str, "exciting")

    messages = MessageFactory()

    something = Event("something", some.object)
    something_exciting = something.when(is_exciting)
    timeline, _ = make_timeline()
    t = timeline.beginning

    timeline.record_event(messages.event("something", "boring"))
    with timeline.frozen():
        timeline.expect_realized(t >> something)
        timeline.expect_not_realized(t >> something_exciting)

    timeline.record_event(messages.event("something", "exciting"))
    with timeline.frozen():
        timeline.expect_realized(t >> something_exciting)


def test_lower_bound(make_timeline):
    timeline, _ = make_timeline()
    timeline.mark("1")
    timeline.mark("2")
    timeline.mark("3")
    timeline.freeze()

    assert (Mark("2") >> (Mark("1") >> Mark("3"))) not in timeline
    assert (Mark("2") >> (Mark("1") & Mark("3"))) not in timeline
    assert (Mark("2") >> (Mark("1") ^ Mark("3"))) in timeline


@pytest.mark.timeout(3)
def test_frozen(make_timeline, daemon):
    timeline, initial_history = make_timeline()
    assert not timeline.is_frozen

    timeline.freeze()
    assert timeline.is_frozen

    timeline.unfreeze()
    assert not timeline.is_frozen

    with timeline.frozen():
        assert timeline.is_frozen
    assert not timeline.is_frozen

    timeline.mark("dum")

    timeline.freeze()
    assert timeline.is_frozen

    worker_started = threading.Event()
    worker_can_proceed = threading.Event()

    @daemon
    def worker():
        worker_started.set()
        worker_can_proceed.wait()
        timeline.mark("dee")

    worker_started.wait()

    assert Mark("dum") in timeline
    assert Mark("dee") not in timeline

    with timeline.unfrozen():
        worker_can_proceed.set()
        worker.join()

    assert Mark("dee") in timeline


@pytest.mark.timeout(3)
def test_unobserved(make_timeline, daemon):
    timeline, initial_history = make_timeline()

    worker_can_proceed = threading.Event()

    @daemon
    def worker():
        messages = MessageFactory()

        worker_can_proceed.wait()
        worker_can_proceed.clear()
        timeline.record_event(messages.event("dum", {}))
        log.debug("dum")

        worker_can_proceed.wait()
        timeline.record_event(messages.event("dee", {}))
        log.debug("dee")

        timeline.record_event(messages.event("dum", {}))
        log.debug("dum")

    timeline.freeze()
    assert timeline.is_frozen

    timeline.proceed()
    assert not timeline.is_frozen
    worker_can_proceed.set()

    dum = timeline.wait_for_next(Event("dum"))
    assert timeline.is_frozen
    assert dum.observed

    # Should be fine since we observed 'dum' by waiting for it.
    timeline.proceed()
    assert not timeline.is_frozen

    worker_can_proceed.set()
    worker.join()

    dum2 = timeline.wait_for_next(Event("dum"), freeze=False)
    assert not timeline.is_frozen
    assert dum2.observed

    timeline.wait_until_realized(Event("dum") >> Event("dum"), freeze=True)
    assert timeline.is_frozen

    dee = dum.next
    assert not dee.observed

    # Should complain since 'dee' is unobserved.
    with pytest.raises(Exception):
        timeline.proceed()

    # Observe it!
    timeline.expect_realized(Event("dee"))
    assert dee.observed

    # Should be good now.
    timeline.proceed()


def test_new(make_timeline):
    timeline, _ = make_timeline()

    m1 = timeline.mark("1")
    timeline.freeze()

    assert timeline.expect_new(Mark("1")) is m1
    with pytest.raises(Exception):
        timeline.expect_new(Mark("2"))

    timeline.proceed()
    m2 = timeline.mark("2")
    timeline.freeze()

    with pytest.raises(Exception):
        timeline.expect_new(Mark("1"))
    assert timeline.expect_new(Mark("2")) is m2
    with pytest.raises(Exception):
        timeline.expect_new(Mark("3"))

    timeline.unfreeze()
    m3 = timeline.mark("3")
    timeline.freeze()

    with pytest.raises(Exception):
        timeline.expect_new(Mark("1"))
    assert timeline.expect_new(Mark("2")) is m2
    assert timeline.expect_new(Mark("3")) is m3

    timeline.proceed()
    m4 = timeline.mark("4")
    timeline.mark("4")
    timeline.freeze()

    with pytest.raises(Exception):
        timeline.expect_new(Mark("1"))
    with pytest.raises(Exception):
        timeline.expect_new(Mark("2"))
    with pytest.raises(Exception):
        timeline.expect_new(Mark("3"))
    assert timeline.expect_new(Mark("4")) is m4


@pytest.mark.timeout(3)
@pytest.mark.parametrize("order", ["mark_then_wait", "wait_then_mark"])
def test_concurrency(make_timeline, daemon, order):
    timeline, initial_history = make_timeline()

    occurrences = []
    worker_can_proceed = threading.Event()

    @daemon
    def worker():
        worker_can_proceed.wait()
        mark = timeline.mark("tada")
        occurrences.append(mark)

    if order == "mark_then_wait":
        worker_can_proceed.set()
        unblock_worker_later = None
    else:

        @daemon
        def unblock_worker_later():
            time.sleep(0.1)
            worker_can_proceed.set()

    mark = timeline.wait_for_next(Mark("tada"), freeze=True)

    worker.join()
    assert mark is occurrences[0]
    assert timeline.last is mark
    assert timeline.history() == initial_history + occurrences

    if unblock_worker_later:
        unblock_worker_later.join()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Helpers to work with Python code.
"""

import py.path
import re

_marked_line_numbers_cache = {}


def get_marked_line_numbers(path):
    """Given a path to a Python source file, extracts line numbers for all lines
    that are marked with #@. For example, given this file::

        print(1)  # @foo
        print(2)
        print(3)  # @bar,baz

    the function will return::

        {"foo": 1, "bar": 3, "baz": 3}
    """

    if isinstance(path, py.path.local):
        path = path.strpath

    try:
        return _marked_line_numbers_cache[path]
    except KeyError:
        pass

    # Read as bytes to avoid decoding errors.
    with open(path, "rb") as f:
        lines = {}
        for i, line in enumerate(f):
            match = re.search(rb"#\s*@(.+?)\s*$", line)
            if match:
                markers = match.group(1).decode("ascii")
                for marker in markers.split(","):
                    lines[marker] = i + 1

    _marked_line_numbers_cache[path] = lines
    return lines



import os
from flask import Flask
from flask import render_template


app = Flask(__name__)
exiting = False


@app.route("/")
def home():
    content = "Flask-Jinja-Test"
    print("break here")  # @bphome
    return render_template("hello.html", title="Hello", content=content)


@app.route("/handled")
def bad_route_handled():
    try:
        raise ArithmeticError("Hello")  # @exc_handled
    except Exception:
        pass
    return render_template("hello.html", title="Hello", content="Flask-Jinja-Test")


@app.route("/unhandled")
def bad_route_unhandled():
    raise ArithmeticError("Hello")  # @exc_unhandled
    return render_template("hello.html", title="Hello", content="Flask-Jinja-Test")


@app.route("/badtemplate")
def bad_template():
    return render_template("bad.html", title="Hello", content="Flask-Jinja-Test")


@app.route("/exit")
def exit_app():
    global exiting
    exiting = True
    return "Done"


@app.teardown_request
def teardown(exception):
    if exiting:
        os._exit(0)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Various means of communication with the debuggee."""

import threading
import socket

from debugpy.common import log, messaging, sockets


class BackChannel(object):
    TIMEOUT = 60

    def __init__(self, session):
        self.session = session
        self.port = None
        self._established = threading.Event()
        self._socket = None
        self._server_socket = None

    def __str__(self):
        return f"BackChannel[{self.session.id}]"

    def listen(self):
        self._server_socket = sockets.create_server("127.0.0.1", 0, self.TIMEOUT)
        _, self.port = self._server_socket.getsockname()
        self._server_socket.listen(0)

        def accept_worker():
            log.info(
                "Listening for incoming connection from {0} on port {1}...",
                self,
                self.port,
            )

            server_socket = self._server_socket
            if server_socket is None:
                return  # concurrent close()

            try:
                sock, _ = server_socket.accept()
            except socket.timeout:
                if self._server_socket is None:
                    return
                else:
                    log.reraise_exception("Timed out waiting for {0} to connect", self)
            except Exception:
                if self._server_socket is None:
                    return
                else:
                    log.reraise_exception("Error accepting connection for {0}:", self)

            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            log.info("Incoming connection from {0} accepted.", self)

            self._socket = sock
            self._setup_stream()

        accept_thread = threading.Thread(target=accept_worker, name=f"{self} listener")
        accept_thread.daemon = True
        accept_thread.start()

    def _setup_stream(self):
        self._stream = messaging.JsonIOStream.from_socket(self._socket, name=str(self))
        self._established.set()

    def receive(self):
        self._established.wait()
        return self._stream.read_json()

    def send(self, value):
        self.session.timeline.unfreeze()
        self._established.wait()
        t = self.session.timeline.mark(("sending", value))
        self._stream.write_json(value)
        return t

    def close(self):
        sock = self._socket
        if sock:
            self._socket = None
            log.debug("Closing {0} client socket...", self)
            try:
                sock.shutdown(socket.SHUT_RDWR)
            except Exception:
                pass
            try:
                sock.close()
            except Exception:
                pass

        server_socket = self._server_socket
        if server_socket:
            self._server_socket = None
            log.debug("Closing {0} server socket...", self)
            try:
                server_socket.shutdown(socket.SHUT_RDWR)
            except Exception:
                pass
            try:
                server_socket.close()
            except Exception:
                pass


class ScratchPad(object):
    def __init__(self, session):
        self.session = session

    def __getitem__(self, key):
        raise NotImplementedError

    def __setitem__(self, key, value):
        """Sets debuggee.scratchpad[key] = value inside the debugged process."""
        log.info("{0} debuggee.scratchpad[{1!r}] = {2!r}", self.session, key, value)
        expr = f"sys.modules['debuggee'].scratchpad[{key!r}] = {value!r}"
        self.session.request("evaluate", {"context": "repl", "expression": expr})



import Cython
from cython_json import source_to_dict
import pytest
import json


def test_dump_ast_error():
    as_dict = source_to_dict("x = [a  10]")
    errors = as_dict["errors"]
    assert len(errors) == 1
    error = errors[0]
    assert error["__node__"] == "CompileError"
    assert error["line"] == 1
    assert error["col"] == 8
    assert "Expected" in error["message_only"]


def test_dump_error():
    contents = """
from distutils import sysconfig
"""
    if isinstance(contents, bytes):
        contents = contents.decode("utf-8")
    source_to_dict(contents)


def test_dump_class():
    contents = """
class A:pass
"""
    if isinstance(contents, bytes):
        contents = contents.decode("utf-8")
    source_to_dict(contents)


def test_comp():
    contents = """
{i: j for i, j in a}
"""
    if isinstance(contents, bytes):
        contents = contents.decode("utf-8")
    source_to_dict(contents)


def test_global():
    contents = """
def method():
  global b
  b = 10
"""
    if isinstance(contents, bytes):
        contents = contents.decode("utf-8")
    source_to_dict(contents)


# def test_dump_custom():
#     with open(r'X:\cython\tests\compile\buildenv.pyx', 'r') as stream:
#         contents = stream.read().decode('utf-8')
#     source_to_dict(contents)


def test_dump_ast():
    data = source_to_dict("x = [a, 10]")
    assert not data["errors"]
    assert data["ast"]["stats"] == [
        {
            "__node__": "SingleAssignment",
            "rhs": {
                "__node__": "List",
                "line": 1,
                "args": [
                    {"__node__": "Name", "line": 1, "col": 5, "name": "a"},
                    {
                        "is_c_literal": "None",
                        "unsigned": "",
                        "value": "10",
                        "constant_result": "10",
                        "__node__": "Int",
                        "line": 1,
                        "type": "long",
                        "col": 8,
                        "longness": "",
                    },
                ],
                "col": 4,
            },
            "lhs": {"__node__": "Name", "line": 1, "col": 0, "name": "x"},
            "line": 1,
            "col": 4,
        }
    ]


if __name__ == "__main__":
    pytest.main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

from tests import debug
from tests.patterns import some


def test_ndarray(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import numpy
        import debuggee

        debuggee.setup()
        a = numpy.array([123, 456], numpy.int32)
        print(a)  # @bp

    with debug.Session() as session:
        session.config["variablePresentation"] = {"all": "hide", "protected": "inline"}
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()
        scopes = session.request("scopes", {"frameId": stop.frame_id})["scopes"]
        globals_ref = scopes[0]["variablesReference"]
        vars = session.request(
            "variables",
            {"variablesReference": globals_ref},
        )["variables"]
        print(vars)

        # Fetch children variables of the array.
        (a,) = (v for v in vars if v["name"] == "a")
        a_vars = session.request(
            "variables",
            {"variablesReference": a["variablesReference"]},
        )["variables"]
        print(a_vars)

        # Fetch the actual array items
        (items,) = (v for v in a_vars if v["name"] == "[0:2] ")
        a_items = session.request(
            "variables",
            {"variablesReference": items["variablesReference"]},
        )["variables"]
        print(a_items)

        assert a_items == [
            some.dict.containing(
                {
                    "type": "int32",
                    "name": "0",
                    "value": some.str.containing("123"),
                    "variablesReference": some.int,
                }
            ),
            some.dict.containing(
                {
                    "type": "int32",
                    "name": "1",
                    "value": some.str.containing("456"),
                    "variablesReference": some.int,
                }
            ),
            some.dict.containing(
                {
                    "type": "int",
                    "name": "len()",
                    "value": "2",
                    "presentationHint": {"attributes": ["readOnly"]},
                    "variablesReference": 0,
                }
            ),
        ]

        session.request_continue()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
import sys
import time

from tests import debug


@pytest.mark.parametrize("count", [1, 3])
def test_thread_count(pyfile, target, run, count):
    @pyfile
    def code_to_debug():
        import debuggee
        import threading
        import time
        import sys

        debuggee.setup()
        stop = False # noqa: F841

        def worker(tid, offset):
            i = 0
            global stop
            while not stop:
                time.sleep(0.01)
                i += 1

        threads = []
        if sys.argv[1] != "1":
            for i in [111, 222]:
                thread = threading.Thread(target=worker, args=(i, len(threads)))
                threads.append(thread)
                thread.start()
        print("check here")  # @bp
        stop = True  # noqa: F841

    with debug.Session() as session:
        with run(session, target(code_to_debug, args=[str(count)])):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop()
        threads = session.request("threads")
        assert len(threads["threads"]) == count
        session.request_continue()


@pytest.mark.parametrize("resume", ["default", "resume_all", "resume_one"])
def test_step_multi_threads(pyfile, target, run, resume):
    @pyfile
    def code_to_debug():
        # After breaking on the thread 1, thread 2 should pause waiting for the event1 to be set,
        # so, when we step return on thread 1, the program should finish if all threads are resumed
        # or should keep waiting for the thread 2 to run if only thread 1 is resumed.

        import debuggee
        import threading

        debuggee.setup()
        event0 = threading.Event()
        event1 = threading.Event()
        event2 = threading.Event()
        event3 = threading.Event()

        def _thread1():
            while not event0.is_set():
                event0.wait(timeout=0.001)
            event1.set()  # @break_thread_1
            while not event2.is_set():
                event2.wait(timeout=0.001)
            # Note: we can only get here if thread 2 is also released.
            event3.set()

        def _thread2():
            event0.set()
            while not event1.is_set():
                event1.wait(timeout=0.001)
            event2.set()
            while not event3.is_set():
                event3.wait(timeout=0.001)

        threads = [
            threading.Thread(target=_thread1, name="thread1"),
            threading.Thread(target=_thread2, name="thread2"),
        ]
        for t in threads:
            t.start()

        for t in threads:
            t.join()

    with debug.Session() as session:
        if resume == "resume_all":
            session.config["steppingResumesAllThreads"] = True
        elif resume == "resume_one":
            session.config["steppingResumesAllThreads"] = False

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()
        threads = session.request("threads")
        assert len(threads["threads"]) == 3

        thread_name_to_id = {t["name"]: t["id"] for t in threads["threads"]}
        assert stop.thread_id == thread_name_to_id["thread1"]

        if resume == "resume_one":
            session.request("stepOut", {"threadId": stop.thread_id})
            # Wait a second and check that threads are still there.
            time.sleep(1)

            stack_trace = session.request(
                "stackTrace", {"threadId": thread_name_to_id["thread1"]}
            )
            assert "_thread1" in [frame["name"] for frame in stack_trace["stackFrames"]]

            stack_trace = session.request(
                "stackTrace", {"threadId": thread_name_to_id["thread2"]}
            )
            assert "_thread2" in [frame["name"] for frame in stack_trace["stackFrames"]]

            session.request_continue()

        else:
            session.request("stepOut", {"threadId": stop.thread_id}, freeze=False)


@pytest.mark.skipif(
    sys.platform not in ["win32", "darwin"] and not sys.platform.startswith("linux"),
    reason="Test not implemented for sys.platform=" + repr(sys.platform),
)
def test_debug_this_thread(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee
        import debugpy
        import sys
        import threading

        debuggee.setup()

        def foo(x):
            debugpy.debug_this_thread()
            event.set()  # @bp
            return 0

        event = threading.Event()

        if sys.platform == "win32":
            from ctypes import CFUNCTYPE, c_void_p, c_size_t, c_uint32, windll

            thread_func_p = CFUNCTYPE(c_uint32, c_void_p)
            thread_func = thread_func_p(
                foo
            )  # must hold a reference to wrapper during the call
            assert windll.kernel32.CreateThread(
                c_void_p(0),
                c_size_t(0),
                thread_func,
                c_void_p(0),
                c_uint32(0),
                c_void_p(0),
            )
        elif sys.platform == "darwin" or sys.platform.startswith("linux"):
            from ctypes import CDLL, CFUNCTYPE, byref, c_void_p, c_ulong
            from ctypes.util import find_library

            libpthread = CDLL(find_library("pthread"))
            thread_func_p = CFUNCTYPE(c_void_p, c_void_p)
            thread_func = thread_func_p(
                foo
            )  # must hold a reference to wrapper during the call
            assert not libpthread.pthread_create(
                byref(c_ulong(0)), c_void_p(0), thread_func, c_void_p(0)
            )
        else:
            pytest.fail(sys.platform)

        event.wait()

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, [code_to_debug.lines["bp"]])

        session.wait_for_stop()
        session.request_continue()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
import sys

from _pydevd_bundle.pydevd_constants import IS_PY312_OR_GREATER
from tests import debug
from tests.debug import runners
from tests.patterns import some


@pytest.mark.parametrize("stop_method", ["breakpoint", "pause"])
@pytest.mark.skipif(IS_PY312_OR_GREATER, reason="Flakey test on 312 and higher")
@pytest.mark.parametrize("is_client_connected", ["is_client_connected", ""])
@pytest.mark.parametrize("wait_for_client", ["wait_for_client", pytest.param("", marks=pytest.mark.skipif(sys.platform.startswith("darwin"), reason="Flakey test on Mac"))])
def test_attach_api(pyfile, wait_for_client, is_client_connected, stop_method):
    @pyfile
    def code_to_debug():
        import debuggee
        import debugpy
        import sys
        import time
        from debuggee import backchannel, scratchpad

        # Test different ways of calling configure(). 
        debugpy.configure(qt="none", subProcess=True, python=sys.executable)
        debugpy.configure({"qt": "none", "subProcess": True, "python": sys.executable})
        debugpy.configure({"qt": "none"}, python=sys.executable)

        debuggee.setup()
        _, host, port, wait_for_client, is_client_connected, stop_method = sys.argv
        port = int(port)
        debugpy.listen(address=(host, port))

        if wait_for_client:
            backchannel.send("wait_for_client")
            debugpy.wait_for_client()

        if is_client_connected:
            backchannel.send("is_client_connected")
            while not debugpy.is_client_connected():
                print("looping until is_client_connected()")
                time.sleep(0.1)

        if stop_method == "breakpoint":
            backchannel.send("breakpoint?")
            assert backchannel.receive() == "proceed"
            debugpy.breakpoint()
            print("break")  # @breakpoint
        else:
            scratchpad["paused"] = False
            backchannel.send("loop?")
            assert backchannel.receive() == "proceed"
            while not scratchpad["paused"]:
                print("looping until paused")
                time.sleep(0.1)

    with debug.Session() as session:
        host, port = runners.attach_connect.host, runners.attach_connect.port
        session.config.update({"connect": {"host": host, "port": port}})

        backchannel = session.open_backchannel()
        session.spawn_debuggee(
            [
                code_to_debug,
                host,
                port,
                wait_for_client,
                is_client_connected,
                stop_method,
            ]
        )
        session.wait_for_adapter_socket()

        session.expect_server_socket()
        session.connect_to_adapter((host, port))
        with session.request_attach():
            pass

        if wait_for_client:
            assert backchannel.receive() == "wait_for_client"

        if is_client_connected:
            assert backchannel.receive() == "is_client_connected"

        if stop_method == "breakpoint":
            assert backchannel.receive() == "breakpoint?"
            backchannel.send("proceed")
            session.wait_for_stop(
                expected_frames=[some.dap.frame(code_to_debug, "breakpoint")]
            )
        elif stop_method == "pause":
            assert backchannel.receive() == "loop?"
            backchannel.send("proceed")
            session.request("pause", freeze=False)
            session.wait_for_stop("pause")
            session.scratchpad["paused"] = True
        else:
            pytest.fail(stop_method)

        session.request_continue()


@pytest.mark.parametrize("run", runners.all_attach_connect)
def test_reattach(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import time
        import debuggee
        import debugpy
        from debuggee import scratchpad

        debuggee.setup()
        debugpy.breakpoint()
        object()  # @first

        scratchpad["exit"] = False
        while not scratchpad["exit"]:
            time.sleep(0.1)
            debugpy.breakpoint()
            object()  # @second

    with debug.Session() as session1:
        session1.captured_output = set()
        session1.expected_exit_code = None  # not expected to exit on disconnect

        with run(session1, target(code_to_debug)):
            expected_adapter_sockets = session1.expected_adapter_sockets.copy()

        session1.wait_for_stop(expected_frames=[some.dap.frame(code_to_debug, "first")])
        session1.disconnect()

    with debug.Session() as session2:
        session2.config.update(session1.config)
        session2.expected_adapter_sockets = expected_adapter_sockets
        if "connect" in session2.config:
            session2.connect_to_adapter(
                (session2.config["connect"]["host"], session2.config["connect"]["port"])
            )

        with session2.request_attach():
            pass

        session2.wait_for_stop(
            expected_frames=[some.dap.frame(code_to_debug, "second")]
        )
        session2.scratchpad["exit"] = True
        session2.request_continue()

    session1.wait_for_exit()


@pytest.mark.parametrize("pid_type", ["int", "str"])
@pytest.mark.skipif(
    not sys.platform.startswith("linux"),
    reason="https://github.com/microsoft/debugpy/issues/311",
)
@pytest.mark.flaky(retries=2, delay=1)
def test_attach_pid_client(pyfile, target, pid_type):
    @pyfile
    def code_to_debug():
        import debuggee
        import time

        debuggee.setup()

        def do_something(i):
            time.sleep(0.2)
            proceed = True
            print(i)  # @bp
            return proceed

        for i in range(500):
            if not do_something(i):
                break

    def before_request(command, arguments):
        if command == "attach":
            assert isinstance(arguments["processId"], int)
            if pid_type == "str":
                arguments["processId"] = str(arguments["processId"])

    session1 = debug.Session()

    session1.before_request = before_request
    session1.config["redirectOutput"] = True

    session1.captured_output = set()
    session1.expected_exit_code = None  # not expected to exit on disconnect

    with session1.attach_pid(target(code_to_debug), wait=False):
        session1.set_breakpoints(code_to_debug, all)

    session1.wait_for_stop(expected_frames=[some.dap.frame(code_to_debug, "bp")])

    pid = session1.config["processId"]

    # Note: don't call session1.disconnect because it'd deadlock in channel.close()
    # (because the fd is in a read() in a different thread, we can't call close() on it).
    session1.request("disconnect")
    session1.wait_for_terminated()

    with debug.Session() as session2:
        with session2.attach_pid(pid, wait=False):
            session2.set_breakpoints(code_to_debug, all)

        stop = session2.wait_for_stop(
            expected_frames=[some.dap.frame(code_to_debug, "bp")]
        )

        # Remove breakpoint and continue.
        session2.set_breakpoints(code_to_debug, [])
        session2.request(
            "setExpression",
            {"frameId": stop.frame_id, "expression": "proceed", "value": "False"},
        )
        session2.scratchpad["exit"] = True
        session2.request_continue()


def test_cancel_wait(pyfile):
    @pyfile
    def code_to_debug():
        import debugpy
        import sys
        import threading
        import time

        from debuggee import backchannel

        def cancel():
            time.sleep(1)
            debugpy.wait_for_client.cancel()

        _, host, port = sys.argv
        port = int(port)
        debugpy.listen(address=(host, port))
        threading.Thread(target=cancel).start()
        debugpy.wait_for_client()
        backchannel.send("exit")

    with debug.Session() as session:
        host, port = runners.attach_connect.host, runners.attach_connect.port
        session.config.update({"connect": {"host": host, "port": port}})
        session.expected_exit_code = None

        backchannel = session.open_backchannel()
        session.spawn_debuggee(
            [
                code_to_debug,
                host,
                port,
            ]
        )

        assert backchannel.receive() == "exit"



import pytest

pytest_plugins = [
    str("_pytest.pytester"),
]


def _run_and_check(testdir_or_pytester, path, check_for="Worked"):
    result = testdir_or_pytester.runpython(path)
    result.stdout.fnmatch_lines([check_for])


if hasattr(pytest, "version_tuple") and pytest.version_tuple[0] >= 7:

    @pytest.fixture
    def testdir_or_pytester(pytester):
        return pytester

else:

    @pytest.fixture
    def testdir_or_pytester(testdir):
        return testdir


def test_run(testdir_or_pytester):
    from tests_python import debugger_unittest
    import sys
    import os

    foo_dir = debugger_unittest._get_debugger_test_file(os.path.join("resources", "launch", "foo"))
    foo_module = "tests_python.resources.launch.foo"

    pydevd_dir = os.path.dirname(os.path.dirname(__file__))
    assert os.path.exists(os.path.join(pydevd_dir, "pydevd.py"))

    _run_and_check(
        testdir_or_pytester,
        testdir_or_pytester.makepyfile(
            """
import sys
sys.path.append(%(pydevd_dir)r)
import pydevd
py_db = pydevd.PyDB()
py_db.ready_to_run = True
py_db.run(%(foo_dir)r)
"""
            % locals()
        ),
    )

    _run_and_check(
        testdir_or_pytester,
        testdir_or_pytester.makepyfile(
            """
import sys
sys.path.append(%(pydevd_dir)r)
import pydevd
py_db = pydevd.PyDB()
py_db.run(%(foo_dir)r, set_trace=False)
"""
            % locals()
        ),
    )

    if sys.version_info[0:2] == (2, 6):
        # Not valid for Python 2.6
        return

    _run_and_check(
        testdir_or_pytester,
        testdir_or_pytester.makepyfile(
            """
import sys
sys.path.append(%(pydevd_dir)r)
sys.argv.append('--as-module')
import pydevd
py_db = pydevd.PyDB()
py_db.ready_to_run = True
py_db.run(%(foo_module)r, is_module=True)
"""
            % locals()
        ),
    )

    _run_and_check(
        testdir_or_pytester,
        testdir_or_pytester.makepyfile(
            """
import sys
sys.argv.append('--as-module')
sys.path.append(%(pydevd_dir)r)
import pydevd
py_db = pydevd.PyDB()
py_db.run(%(foo_module)r, is_module=True, set_trace=False)
"""
            % locals()
        ),
    )


def test_run_on_local_module_without_adding_to_pythonpath(testdir_or_pytester):
    import sys
    import os

    pydevd_dir = os.path.dirname(os.path.dirname(__file__))
    assert os.path.exists(os.path.join(pydevd_dir, "pydevd.py"))

    foo_module = "local_foo"
    with open(os.path.join(os.getcwd(), "local_foo.py"), "w") as stream:
        stream.write('print("WorkedLocalFoo")')

    _run_and_check(
        testdir_or_pytester,
        testdir_or_pytester.makepyfile(
            """
import sys
import os
sys.path.append(%(pydevd_dir)r)
sys.argv.append('--as-module')
cwd = os.path.abspath(os.getcwd())
while cwd in sys.path:
    sys.path.remove(cwd)
import pydevd
py_db = pydevd.PyDB()
py_db.ready_to_run = True
py_db.run(%(foo_module)r, is_module=True)
"""
            % locals()
        ),
        check_for="WorkedLocalFoo",
    )

    _run_and_check(
        testdir_or_pytester,
        testdir_or_pytester.makepyfile(
            """
import sys
import os
sys.argv.append('--as-module')
sys.path.append(%(pydevd_dir)r)
cwd = os.path.abspath(os.getcwd())
while cwd in sys.path:
    sys.path.remove(cwd)
import pydevd
py_db = pydevd.PyDB()
py_db.run(%(foo_module)r, is_module=True, set_trace=False)
"""
            % locals()
        ),
        check_for="WorkedLocalFoo",
    )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
import sys

from _pydevd_bundle.pydevd_constants import IS_PY312_OR_GREATER
from tests import debug
from tests.debug import runners, targets
from tests.patterns import some
from tests.timeline import Event

str_matching_ArithmeticError = some.str.matching(r"(.+\.)?ArithmeticError")


@pytest.mark.parametrize("raised", ["raised", ""])
@pytest.mark.parametrize("uncaught", ["uncaught", ""])
def test_vsc_exception_options_raise_with_except(pyfile, target, run, raised, uncaught):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        def raise_with_except():
            try:
                raise ArithmeticError("bad code")  # @exc
            except Exception:
                pass

        raise_with_except()

    with debug.Session() as session:
        session.expected_exit_code = some.int
        with run(session, target(code_to_debug)):
            session.request(
                "setExceptionBreakpoints", {"filters": list({raised, uncaught} - {""})}
            )

        expected = some.dict.containing(
            {
                "exceptionId": str_matching_ArithmeticError,
                "description": "bad code",
                "breakMode": "always" if raised else "unhandled",
                "details": some.dict.containing(
                    {
                        "typeName": str_matching_ArithmeticError,
                        "message": "bad code",
                        "source": some.path(code_to_debug),
                    }
                ),
            }
        )

        if raised:
            stop = session.wait_for_stop(
                "exception",
                expected_text=str_matching_ArithmeticError,
                expected_description="bad code",
                expected_frames=[some.dap.frame(code_to_debug, line="exc")],
            )
            exc_info = session.request("exceptionInfo", {"threadId": stop.thread_id})
            assert exc_info == expected
            session.request_continue()

        if uncaught:
            # Exception is caught by try..except, so there should be no stop.
            pass


@pytest.mark.parametrize("raised", ["raised", ""])
@pytest.mark.parametrize("uncaught", ["uncaught", ""])
def test_vsc_exception_options_raise_without_except(
    pyfile, target, run, raised, uncaught
):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        def raise_without_except():
            raise ArithmeticError("bad code")  # @exc

        raise_without_except()

    with debug.Session() as session:
        session.ignore_unobserved.append(Event("stopped"))
        session.expected_exit_code = some.int

        with run(session, target(code_to_debug)):
            session.request(
                "setExceptionBreakpoints", {"filters": list({raised, uncaught} - {""})}
            )

        expected_exc_info = some.dict.containing(
            {
                "exceptionId": str_matching_ArithmeticError,
                "description": "bad code",
                "breakMode": "always" if raised else "unhandled",
                "details": some.dict.containing(
                    {
                        "typeName": str_matching_ArithmeticError,
                        "message": "bad code",
                        "source": some.path(code_to_debug),
                    }
                ),
            }
        )

        if raised:
            stop = session.wait_for_stop(
                "exception", expected_frames=[some.dap.frame(code_to_debug, line="exc")]
            )
            exc_info = session.request("exceptionInfo", {"threadId": stop.thread_id})
            assert expected_exc_info == exc_info
            session.request_continue()

            # NOTE: debugger stops at each frame if raised and is uncaught
            # This behavior can be changed by updating 'notify_on_handled_exceptions'
            # setting we send to pydevd to notify only once. In our test code, we have
            # two frames, hence two stops.
            session.wait_for_stop("exception")
            session.request_continue()

        if uncaught:
            stop = session.wait_for_stop(
                "exception", expected_frames=[some.dap.frame(code_to_debug, line="exc")]
            )
            expected_exc_info = some.dict.containing(
                {
                    "exceptionId": str_matching_ArithmeticError,
                    "description": "bad code",
                    "breakMode": "unhandled",  # Only difference from previous expected is breakMode.
                    "details": some.dict.containing(
                        {
                            "typeName": str_matching_ArithmeticError,
                            "message": "bad code",
                            "source": some.path(code_to_debug),
                        }
                    ),
                }
            )
            exc_info = session.request("exceptionInfo", {"threadId": stop.thread_id})
            assert expected_exc_info == exc_info
            session.request_continue()


@pytest.mark.skipif(
    sys.platform == "darwin",
    reason="https://github.com/microsoft/ptvsd/issues/1988",
)
@pytest.mark.flaky(retries=2, delay=1)
@pytest.mark.parametrize("target", targets.all_named)
@pytest.mark.parametrize("run", runners.all)
@pytest.mark.parametrize("raised", ["raised", ""])
@pytest.mark.parametrize("uncaught", ["uncaught", ""])
@pytest.mark.parametrize("zero", ["zero", ""])
@pytest.mark.parametrize("exit_code", [0, 1, "nan"])
def test_systemexit(pyfile, target, run, raised, uncaught, zero, exit_code):
    @pyfile
    def code_to_debug():
        import debuggee
        import sys

        debuggee.setup()
        exit_code = eval(sys.argv[1])
        print("sys.exit(%r)" % (exit_code,))
        try:
            sys.exit(exit_code)  # @handled
        except SystemExit:
            pass
        sys.exit(exit_code)  # @unhandled

    filters = []
    if raised:
        filters += ["raised"]
    if uncaught:
        filters += ["uncaught"]

    with debug.Session() as session:
        session.expected_exit_code = some.int
        session.config["breakOnSystemExitZero"] = bool(zero)

        with run(session, target(code_to_debug, args=[repr(exit_code)])):
            session.request("setExceptionBreakpoints", {"filters": filters})

        # When breaking on raised exceptions, we'll stop on both lines,
        # unless it's SystemExit(0) and we asked to ignore that.
        if raised and (zero or exit_code != 0):
            session.wait_for_stop(
                "exception",
                expected_frames=[some.dap.frame(code_to_debug, line="handled")],
            )
            session.request_continue()

            session.wait_for_stop(
                "exception",
                expected_frames=[some.dap.frame(code_to_debug, line="unhandled")],
            )
            session.request_continue()

        # When breaking on uncaught exceptions, we'll stop on the second line,
        # unless it's SystemExit(0) and we asked to ignore that.
        # Note that if both raised and uncaught filters are set, there will be
        # two stop for the second line - one for exception being raised, and one
        # for it unwinding the stack without finding a handler. The block above
        # takes care of the first stop, so here we just take care of the second.
        if uncaught and (zero or exit_code != 0):
            session.wait_for_stop(
                "exception",
                expected_frames=[some.dap.frame(code_to_debug, line="unhandled")],
            )
            session.request_continue()


@pytest.mark.parametrize(
    "break_mode", ["always", "never", "unhandled", "userUnhandled"]
)
@pytest.mark.parametrize(
    "exceptions",
    [
        ["RuntimeError"],
        ["AssertionError"],
        ["RuntimeError", "AssertionError"],
        [],  # Add the whole Python Exceptions category.
    ],
)
def test_raise_exception_options(pyfile, target, run, exceptions, break_mode):
    if break_mode in ("never", "unhandled", "userUnhandled"):
        expect_exceptions = []
        if break_mode != "never" and (not exceptions or "AssertionError" in exceptions):
            # Only AssertionError is raised in this use-case.
            expect_exceptions = ["AssertionError"]

        @pyfile
        def code_to_debug():
            import debuggee

            debuggee.setup()
            raise AssertionError()  # @AssertionError

    else:
        expect_exceptions = exceptions[:]
        if not expect_exceptions:
            # Deal with the Python Exceptions category
            expect_exceptions = ["RuntimeError", "AssertionError", "IndexError"]

        @pyfile
        def code_to_debug():
            import debuggee

            debuggee.setup()

            try:
                raise RuntimeError()  # @RuntimeError
            except RuntimeError:
                pass
            try:
                raise AssertionError()  # @AssertionError
            except AssertionError:
                pass
            try:
                raise IndexError()  # @IndexError
            except IndexError:
                pass

    with debug.Session() as session:
        session.ignore_unobserved.append(Event("stopped"))
        session.expected_exit_code = some.int

        with run(session, target(code_to_debug)):
            path = [{"names": ["Python Exceptions"]}]
            if exceptions:
                path.append({"names": exceptions})
            session.request(
                "setExceptionBreakpoints",
                {
                    "filters": [],  # Unused when exceptionOptions is passed.
                    "exceptionOptions": [
                        {
                            "path": path,
                            "breakMode": break_mode,  # Can be "never", "always", "unhandled", "userUnhandled"
                        }
                    ],
                },
            )

        for expected_exception in expect_exceptions:
            session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(code_to_debug, line=expected_exception)
                ],
            )
            session.request_continue()


@pytest.mark.parametrize("target", targets.all_named)
@pytest.mark.parametrize("run", runners.all)
@pytest.mark.parametrize("exit_code", [0, 3])
@pytest.mark.parametrize("break_on_system_exit_zero", ["break_on_system_exit_zero", ""])
@pytest.mark.parametrize("django", ["django", ""])
@pytest.mark.skipif(sys.platform == 'win32' and IS_PY312_OR_GREATER, reason="Flakey test")
def test_success_exitcodes(
    pyfile, target, run, exit_code, break_on_system_exit_zero, django
):
    @pyfile
    def code_to_debug():
        import debuggee
        import sys

        debuggee.setup()
        exit_code = eval(sys.argv[1])
        print("sys.exit(%r)" % (exit_code,))
        sys.exit(exit_code)

    with debug.Session() as session:
        session.expected_exit_code = some.int
        session.config["breakOnSystemExitZero"] = bool(break_on_system_exit_zero)
        session.config["django"] = bool(django)

        with run(session, target(code_to_debug, args=[repr(exit_code)])):
            session.request("setExceptionBreakpoints", {"filters": ["uncaught"]})

        if break_on_system_exit_zero or (not django and exit_code == 3):
            # If "breakOnSystemExitZero" was specified, we should always break.
            # Otherwise, we should not break if the exit code indicates successful
            # exit. 0 always indicates success, and 3 indicates failure only if
            # Django debugging wasn't enabled.
            session.wait_for_stop("exception")
            session.request_continue()


@pytest.mark.parametrize("max_frames", ["default", "all", 10])
def test_exception_stack(pyfile, target, run, max_frames):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        def do_something(n):
            if n <= 0:
                raise ArithmeticError("bad code")  # @unhandled
            do_something2(n - 1)

        def do_something2(n):
            do_something(n - 1)

        do_something(100)

    with debug.Session() as session:
        session.expected_exit_code = some.int

        max_frames, (min_expected_lines, max_expected_lines) = {
            "all": (0, (100, 308)),
            "default": (None, (100, 308)),
            10: (10, (10, 32)),
        }[max_frames]
        if max_frames is not None:
            session.config["maxExceptionStackFrames"] = max_frames

        with run(session, target(code_to_debug)):
            session.request("setExceptionBreakpoints", {"filters": ["uncaught"]})

        stop = session.wait_for_stop(
            "exception",
            expected_frames=[some.dap.frame(code_to_debug, line="unhandled")],
        )
        exc_info = session.request("exceptionInfo", {"threadId": stop.thread_id})
        expected_exc_info = some.dict.containing(
            {
                "exceptionId": str_matching_ArithmeticError,
                "description": "bad code",
                "breakMode": "unhandled",
                "details": some.dict.containing(
                    {
                        "typeName": str_matching_ArithmeticError,
                        "message": "bad code",
                        "source": some.path(code_to_debug),
                    }
                ),
            }
        )
        assert expected_exc_info == exc_info
        stack_str = exc_info["details"]["stackTrace"]
        stack_line_count = len(stack_str.split("\n"))
        assert min_expected_lines <= stack_line_count <= max_expected_lines

        session.request_continue()



"""
@author Fabio Zadrozny
"""

from _pydev_bundle import _pydev_imports_tipper
import inspect
import pytest
import sys
import unittest

try:
    import __builtin__  # @UnusedImport

    BUILTIN_MOD = "__builtin__"
except ImportError:
    BUILTIN_MOD = "builtins"

IS_JYTHON = sys.platform.find("java") != -1

HAS_WX = False


@pytest.mark.skipif(IS_JYTHON, reason="CPython related test")
class TestCPython(unittest.TestCase):
    def p(self, t):
        for a in t:
            sys.stdout.write("%s\n" % (a,))

    def test_imports3(self):
        tip = _pydev_imports_tipper.generate_tip("os")
        ret = self.assert_in("path", tip)
        self.assertEqual("", ret[2])

    def test_imports2(self):
        try:
            tip = _pydev_imports_tipper.generate_tip("OpenGL.GLUT")
            self.assert_in("glutDisplayFunc", tip)
            self.assert_in("glutInitDisplayMode", tip)
        except ImportError:
            pass

    def test_imports4(self):
        try:
            tip = _pydev_imports_tipper.generate_tip("mx.DateTime.mxDateTime.mxDateTime")
            self.assert_in("now", tip)
        except ImportError:
            pass

    def test_imports5(self):
        tip = _pydev_imports_tipper.generate_tip("%s.list" % BUILTIN_MOD)
        s = self.assert_in("sort", tip)
        self.check_args(
            s,
            "(cmp=None, key=None, reverse=False)",
            "(self, object cmp, object key, bool reverse)",
            "(self, cmp: object, key: object, reverse: bool)",
            "(key=None, reverse=False)",
            "(self, key=None, reverse=False)",
            "(self, cmp, key, reverse)",
            "(self, key, reverse)",
        )

    def test_imports2a(self):
        tips = _pydev_imports_tipper.generate_tip("%s.RuntimeError" % BUILTIN_MOD)
        self.assert_in("__doc__", tips)

    def test_imports2b(self):
        try:
            file
        except:
            pass
        else:
            tips = _pydev_imports_tipper.generate_tip("%s" % BUILTIN_MOD)
            t = self.assert_in("file", tips)
            self.assertTrue("->" in t[1].strip() or "file" in t[1])

    def test_imports2c(self):
        try:
            file  # file is not available on py 3
        except:
            pass
        else:
            tips = _pydev_imports_tipper.generate_tip("%s.file" % BUILTIN_MOD)
            t = self.assert_in("readlines", tips)
            self.assertTrue("->" in t[1] or "sizehint" in t[1])

    def test_imports(self):
        """
        You can print_ the results to check...
        """
        if HAS_WX:
            tip = _pydev_imports_tipper.generate_tip("wxPython.wx")
            self.assert_in("wxApp", tip)

            tip = _pydev_imports_tipper.generate_tip("wxPython.wx.wxApp")

            try:
                tip = _pydev_imports_tipper.generate_tip("qt")
                self.assert_in("QWidget", tip)
                self.assert_in("QDialog", tip)

                tip = _pydev_imports_tipper.generate_tip("qt.QWidget")
                self.assert_in("rect", tip)
                self.assert_in("rect", tip)
                self.assert_in("AltButton", tip)

                tip = _pydev_imports_tipper.generate_tip("qt.QWidget.AltButton")
                self.assert_in("__xor__", tip)

                tip = _pydev_imports_tipper.generate_tip("qt.QWidget.AltButton.__xor__")
                self.assert_in("__class__", tip)
            except ImportError:
                pass

        tip = _pydev_imports_tipper.generate_tip(BUILTIN_MOD)
        #        for t in tip[1]:
        #            print_ t
        self.assert_in("object", tip)
        self.assert_in("tuple", tip)
        self.assert_in("list", tip)
        self.assert_in("RuntimeError", tip)
        self.assert_in("RuntimeWarning", tip)

        # Remove cmp as it's not available on py 3
        # t = self.assert_in('cmp' , tip)
        # self.check_args(t, '(x, y)', '(object x, object y)', '(x: object, y: object)') #args

        t = self.assert_in("isinstance", tip)
        self.check_args(
            t,
            "(object, class_or_type_or_tuple)",
            "(object o, type typeinfo)",
            "(o: object, typeinfo: type)",
            "(obj, class_or_tuple)",
            "(obj, klass_or_tuple)",
        )  # args

        t = self.assert_in("compile", tip)
        self.check_args(
            t,
            "(source, filename, mode)",
            "()",
            "(o: object, name: str, val: object)",
            "(source, filename, mode, flags, dont_inherit, optimize)",
            "(source, filename, mode, flags, dont_inherit)",
            "(source, filename, mode, flags, dont_inherit, optimize, _feature_version=-1)",
            "(source, filename, mode, flags, dont_inherit, optimize, _feature_version)",
        )  # args

        t = self.assert_in("setattr", tip)
        self.check_args(
            t,
            "(object, name, value)",
            "(object o, str name, object val)",
            "(o: object, name: str, val: object)",
            "(obj, name, value)",
            "(object, name, val)",
        )  # args

        try:
            import compiler

            compiler_module = "compiler"
        except ImportError:
            try:
                import ast

                compiler_module = "ast"
            except ImportError:
                compiler_module = None

        if compiler_module is not None:  # Not available in iron python
            tip = _pydev_imports_tipper.generate_tip(compiler_module)
            if compiler_module == "compiler":
                self.assert_args("parse", "(buf, mode)", tip)
                self.assert_args("walk", "(tree, visitor, walker, verbose)", tip)
                self.assert_in("parseFile", tip)
            else:
                self.assert_args(
                    "parse",
                    [
                        "(source, filename, mode)",
                        "(source, filename, mode, type_comments=False, feature_version=None)",
                        "(source, filename, mode, type_comments=False, feature_version=None, optimize=-1)",
                    ],
                    tip,
                )
                self.assert_args("walk", "(node)", tip)
            self.assert_in("parse", tip)

    def check_args(self, t, *expected):
        for x in expected:
            if x == t[2]:
                return
        self.fail("Found: %s. Expected: %s" % (t[2], expected))

    def assert_args(self, tok, args, tips):
        if not isinstance(args, (list, tuple)):
            args = (args,)

        for a in tips[1]:
            if tok == a[0]:
                for arg in args:
                    if arg == a[2]:
                        return
                raise AssertionError("%s not in %s", a[2], args)

        raise AssertionError("%s not in %s", tok, tips)

    def assert_in(self, tok, tips):
        for a in tips[1]:
            if tok == a[0]:
                return a
        raise AssertionError("%s not in %s" % (tok, tips))

    def test_search(self):
        s = _pydev_imports_tipper.search_definition("inspect.ismodule")
        (f, line, col), foundAs = s
        self.assertTrue(line > 0)

    def test_dot_net_libraries(self):
        if sys.platform == "cli":
            tip = _pydev_imports_tipper.generate_tip("System.Drawing")
            self.assert_in("Brushes", tip)

            tip = _pydev_imports_tipper.generate_tip("System.Drawing.Brushes")
            self.assert_in("Aqua", tip)

    def test_tips_hasattr_failure(self):
        class MyClass(object):
            def __getattribute__(self, attr):
                raise RuntimeError()

        obj = MyClass()

        _pydev_imports_tipper.generate_imports_tip_for_module(obj)

    def test_inspect(self):
        class C(object):
            def metA(self, a, b):
                pass

        obj = C.metA
        if inspect.ismethod(obj):
            pass


#            print_ obj.im_func
#            print_ inspect.getargspec(obj.im_func)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import os
import re
import threading

from debugpy.common import log


class CapturedOutput(object):
    """Captures stdout and stderr of the debugged process."""

    def __init__(self, session, **fds):
        self.session = session
        self._lock = threading.Lock()
        self._chunks = {}
        self._worker_threads = []

        for stream_name, fd in fds.items():
            log.info("Capturing {0} {1}", session.debuggee_id, stream_name)
            self._capture(fd, stream_name)

    def __str__(self):
        return f"CapturedOutput[{self.session.id}]"

    def _worker(self, fd, name):
        chunks = self._chunks[name]
        try:
            while True:
                try:
                    chunk = os.read(fd, 0x1000)
                except Exception:
                    break
                if not len(chunk):
                    break

                lines = "\n".join(
                    repr(line) for line, _ in re.findall(b"(.+?(\n|$))", chunk)
                )
                log.info("{0} {1}:\n{2}", self.session.debuggee_id, name, lines)

                with self._lock:
                    chunks.append(chunk)
        finally:
            os.close(fd)

    def _capture(self, fd, name):
        assert name not in self._chunks
        self._chunks[name] = []

        thread = threading.Thread(
            target=lambda: self._worker(fd, name), name=f"{self} {name}"
        )
        thread.daemon = True
        thread.start()
        self._worker_threads.append(thread)

    def wait(self, timeout=None):
        """Wait for all remaining output to be captured."""
        if not self._worker_threads:
            return
        log.debug("Waiting for remaining {0} output...", self.session.debuggee_id)
        for t in self._worker_threads:
            t.join(timeout)
        self._worker_threads[:] = []

    def _output(self, which, encoding, lines):
        try:
            result = self._chunks[which]
        except KeyError:
            raise AssertionError(
                f"{which} was not captured for {self.session.debuggee_id}"
            )

        with self._lock:
            result = b"".join(result)
        if encoding is not None:
            result = result.decode(encoding)

        return result.splitlines() if lines else result

    def stdout(self, encoding=None):
        """Returns stdout captured from the debugged process, as a single string.

        If encoding is None, returns bytes. Otherwise, returns str.
        """
        return self._output("stdout", encoding, lines=False)

    def stderr(self, encoding=None):
        """Returns stderr captured from the debugged process, as a single string.

        If encoding is None, returns bytes. Otherwise, returns str.
        """
        return self._output("stderr", encoding, lines=False)

    def stdout_lines(self, encoding=None):
        """Returns stdout captured from the debugged process, as a list of lines.

        If encoding is None, each line is bytes. Otherwise, each line is str.
        """
        return self._output("stdout", encoding, lines=True)

    def stderr_lines(self, encoding=None):
        """Returns stderr captured from the debugged process, as a list of lines.

        If encoding is None, each line is bytes. Otherwise, each line is str.
        """
        return self._output("stderr", encoding, lines=True)



import asyncio
import sys


async def gen():
    f = sys._getframe()
    for i in range(10):
        await asyncio.sleep(.01)
        assert f is sys._getframe()
        yield i


async def run():
    async for p in gen():
        print(p)


if __name__ == "__main__":
    loop = asyncio.get_event_loop_policy().get_event_loop()
    loop.run_until_complete(run())
    print('TEST SUCEEDED')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
import sys

from _pydevd_bundle.pydevd_constants import IS_PY312_OR_GREATER
from tests import debug
from tests.debug import runners

# When debuggee exits, there's no guarantee currently that all "output" events have
# already been sent. To ensure that they are, all tests below must set a breakpoint
# on the last line of the debuggee, and stop on it. Since debugger sends its events
# sequentially, by the time we get to "stopped", we also have all the output events.


@pytest.mark.parametrize("run", runners.all)
def test_with_no_output(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        x = 4  # @wait_for_output

    with debug.Session() as session:
        session.config["redirectOutput"] = True

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop("breakpoint")
        session.request_continue()

    output = session.output("stdout")
    lines = []
    for line in output.splitlines(keepends=True):
        if not line.startswith("Attaching to PID:"):
            lines.append(line)
    
    output = "".join(lines)
    
    assert not output
    assert not session.output("stderr")
    if session.debuggee is not None:
        assert not session.captured_stdout()
        assert not session.captured_stderr()


def test_with_tab_in_output(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        a = "\t".join(("Hello", "World"))
        print(a)
        x = 4  # @wait_for_output

    with debug.Session() as session:
        session.config["redirectOutput"] = True

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop()
        session.request_continue()

    assert session.output("stdout").startswith("Hello\tWorld")


@pytest.mark.parametrize("redirect_mode", ["internalConsole", "redirectOutput"])
def test_redirect_output_and_eval(pyfile, target, run, redirect_mode):
    @pyfile
    def code_to_debug():
        import debuggee
        import sys

        debuggee.setup()
        sys.stdout.write("line\n")
        x = 4  # @wait_for_output

    with debug.Session() as session:
        if redirect_mode == "redirectOutput":
            session.config["redirectOutput"] = True
        elif redirect_mode == "internalConsole":
            session.config["console"] = "internalConsole"
        else:
            raise AssertionError("Unexpected: " + redirect_mode)

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        stop = session.wait_for_stop()
        session.request(
            "evaluate",
            {
                "expression": "sys.stdout.write('evaluated\\n')",
                "frameId": stop.frame_id,
                "context": "repl",
            },
        )

        session.request_continue()

    assert "line" in session.output("stdout")
    assert "evaluated" in session.output("stdout")


@pytest.mark.parametrize("run", runners.all)
@pytest.mark.parametrize("redirect", ["enabled", "disabled"])
@pytest.mark.skipif(IS_PY312_OR_GREATER, reason="Flakey test")
def test_redirect_output(pyfile, target, run, redirect):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        for i in [111, 222, 333, 444]:
            print(i)

        x = 4  # @wait_for_output

    with debug.Session() as session:
        session.config["redirectOutput"] = redirect == "enabled"

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop()
        session.request_continue()

    output = session.output("stdout")
    lines = []
    for line in output.splitlines(keepends=True):
        if not line.startswith("Attaching to PID:"):
            lines.append(line)
    
    output = "".join(lines)
    if redirect == "enabled":
        assert output == "111\n222\n333\n444\n"
    else:
        assert not output


def test_non_ascii_output(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee
        import sys

        debuggee.setup()
        a = b"\xc3\xa9 \xc3\xa0 \xc3\xb6 \xc3\xb9\n"
        sys.stdout.buffer.write(a)
        x = 4  # @wait_for_output

    with debug.Session() as session:
        session.config["redirectOutput"] = True

        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop()
        session.request_continue()

    output = session.output("stdout").encode("utf-8", "replace")

    assert output in (
        b"\xc3\xa9 \xc3\xa0 \xc3\xb6 \xc3\xb9\n",
        b"\xc3\x83\xc2\xa9 \xc3\x83\xc2\xa0 \xc3\x83\xc2\xb6 \xc3\x83\xc2\xb9\n",
    )


if sys.platform == "win32":

    @pytest.mark.parametrize("redirect_output", ["", "redirect_output"])
    def test_pythonw_output(pyfile, target, run, redirect_output):
        @pyfile
        def code_to_debug():
            import debuggee

            debuggee.setup()
            print("ok")
            x = 4  # @wait_for_output

        with debug.Session() as session:
            # Don't capture launcher output - we want to see how it handles not
            # having sys.stdin and sys.stdout available.
            session.captured_output = set()

            session.config["pythonPath"] = sys.executable + "/../pythonw.exe"
            session.config["redirectOutput"] = bool(redirect_output)

            with run(session, target(code_to_debug)):
                session.set_breakpoints(code_to_debug, all)

            session.wait_for_stop()
            session.request_continue()

        assert session.output("stdout") == ("ok\n" if redirect_output else "")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Do not import this package directly - import tests.patterns.some instead.
"""

# Wire up some.dap to be an alias for dap, to allow writing some.dap.id etc.
from tests.patterns import some
from tests.patterns import dap

some.dap = dap



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""A watchdog process for debuggee processes spawned by tests.

Interacts with the main test runner process over stdio, and keeps track of running
debugpy processes. If the test runner process goes down, any debugpy test processes
are automatically killed.
"""

__all__ = ["start", "register_spawn", "unregister_spawn"]

import atexit
import os
import psutil
import subprocess
import sys
import threading
import time

from debugpy.common import log, messaging
from tests.watchdog import worker

WATCHDOG_TIMEOUT = 10


_name = f"watchdog-{os.getpid()}"
_stream = None
_process = None
_worker_log_filename = None


def start():
    global _stream, _process, _worker_log_filename
    if _stream is not None:
        return
    if sys.version_info >= (3, 13):
        return

    args = [sys.executable, worker.__file__, str(os.getpid())]
    log.info(
        "Spawning {0} for tests-{1}:\n\n{2}",
        _name,
        os.getpid(),
        "\n".join(repr(s) for s in args),
    )

    _process = psutil.Popen(
        args, bufsize=0, stdin=subprocess.PIPE, stdout=subprocess.PIPE
    )

    _stream = messaging.JsonIOStream(_process.stdout, _process.stdin, _name)

    event, _worker_log_filename = _stream.read_json()
    assert event == "watchdog"

    atexit.register(stop)


def _dump_worker_log(command, problem, exc_info=None):
    reason = f"{_name}.{command}() {problem}"
    if _worker_log_filename is None:
        reason += ", but there is no log."
    else:
        try:
            with open(_worker_log_filename) as f:
                worker_log = f.read()
        except Exception:
            reason += f", but log {_worker_log_filename} could not be retrieved."
        else:
            reason += f"; watchdog worker process log:\n\n{worker_log}"

    if exc_info is None:
        log.error("{0}", reason)
    else:
        log.swallow_exception("{0}", reason, exc_info=exc_info)
    return reason


def _invoke(command, *args):
    def timeout():
        time.sleep(WATCHDOG_TIMEOUT)
        if timeout.occurred is None:
            reason = _dump_worker_log(command, "timed out")
            timeout.occurred = reason

    timeout.occurred = None
    timeout_thread = threading.Thread(target=timeout)
    timeout_thread.daemon = True
    timeout_thread.start()
    try:
        try:
            _stream.write_json([command] + list(args))
            response = _stream.read_json()
            assert response == ["ok"], f"{_name} {response!r}"
        finally:
            timeout.occurred = False
    except Exception:
        _dump_worker_log(command, "failed", sys.exc_info())
        raise
    else:
        assert not timeout.occurred, str(timeout.occurred)


def stop():
    if _stream is None:
        return

    try:
        _invoke("stop")
        _stream.close()
    except Exception:
        log.swallow_exception()


def register_spawn(pid, name):
    if sys.version_info >= (3, 13):
        return
    if _stream is None:
        start()
    _invoke("register_spawn", pid, name)


def unregister_spawn(pid, name):
    if sys.version_info >= (3, 13):
        return
    assert _stream is not None
    _invoke("unregister_spawn", pid, name)
